from __future__ import annotations  # For type hinting within class definitions
import numpy.typing as npt
import numpy as np
import abc
import warnings
import datetime
from scipy import optimize as sopt
from types import NoneType
from typing import Self, overload, Literal, override, TYPE_CHECKING
from pyNexafs.nexafs.scan import scanBase, scanAbstract
from pyNexafs.nexafs.normalisation.norm_settings import (
    # Enumerates
    normMethod,
    extSelection,
    edgeNormPre,
    edgeNormPost,
    # Config Classes
    configBase,
    configChannel,
    configExternalChannel,
    configEdges,
)

if TYPE_CHECKING:
    from pyNexafs.parsers import parserBase  # For type hinting only.


class scanAbstractNorm(scanAbstract, metaclass=abc.ABCMeta):
    """
    Abstract class to define the common methods used in all normalized scans.

    Parameters
    ----------
    scan : scan_abstract
        The original scan object to apply the normalisation to.
    apply_to : list[str | int] | None, optional
        The list of y-channel labels (or indexes) to apply the normalisation to.
        Can also be a list of integers, representing the index of the y-channel, which
        will be converted to a list of strings to reduce ambiguity if labels are available.
        By default None (applies to all channels).
    conv_indexes : bool, optional
        Whether to convert integer indexes in `apply_to` to string labels if available.
        Default is True.
    """

    def __init__(
        self,
        scan: scanAbstract,
        apply_to: list[int | str] | None = None,
        conv_indexes: bool = True,
    ) -> None:
        self._origin = scan
        """The original scan object."""
        self._config = None
        """The normalisation configuration object, used to save settings."""
        self._apply_to: list[str | int] | None
        """The list of y-channel labels (or indexes) to apply the normalisation to."""

        # Set the apply_to attribute
        y_labels = scan.y_labels
        if apply_to is not None:
            if all(isinstance(i, str) for i in apply_to):
                # Create a copy
                self._apply_to = apply_to.copy()  # type: ignore - all i are str.
                return
            if y_labels is not None:
                if all(isinstance(i, (int, str)) for i in apply_to):
                    # Convert indexes to labels when available
                    self._apply_to = []
                    for i in range(len(apply_to)):
                        item = apply_to[i]
                        if conv_indexes and isinstance(item, int):
                            label = y_labels[item]
                            self._apply_to.append(
                                i if label is None else label
                            )  # Use the index
                        else:
                            # not conv_indexes and int, or string
                            self._apply_to.append(item)
                    return
                else:
                    raise ValueError(
                        "`apply_to` must be a list of integers or a list of strings."
                    )
            else:
                if all(isinstance(i, int) for i in apply_to):
                    # No labels available, use indexes
                    self._apply_to = apply_to.copy()  # type: ignore - all i are int.
                    return
                else:
                    raise ValueError(
                        "Y labels are not defined in the scan object. `apply_to` must be a list of integers, but some elements are strings."
                    )
        else:
            # Assign the apply_to attribute
            self._apply_to = None

    def root_parser(self) -> "parserBase":
        """
        Recursively collect the original parser.

        This method traverses through any chained normalisation objects to find the root scan.

        Returns
        -------
        scan_abstract
            The original scan object.

        Raises
        ------
        ValueError
            If the root scan does not have a parser associated with it.
        """
        root_scan = self.root_scan()
        assert isinstance(root_scan, scanBase), (
            f"Root scan is not a `scanBase` instance (instead {type(root_scan)}), which is required to access the parser."
        )
        if root_scan.parser is None:
            raise ValueError("The root scan does not have a parser associated with it.")
        return root_scan.parser

    def root_scan(self) -> scanAbstract:
        """
        Recursively collect the original parser.

        This method traverses through any chained normalisation objects to find the root scan.

        Returns
        -------
        scan_abstract
            The original scan object.
        """
        if isinstance(self._origin, scanAbstractNorm):
            return self._origin.root_scan()
        return self._origin

    @override
    @abc.abstractmethod
    def copy(self) -> scanAbstractNorm:
        """
        Copy the scan object.

        Returns
        -------
        scan_abstract
            A copy of the scan object.
        """
        copy_obj = self.__class__(scan=self._origin, apply_to=self._apply_to)
        return copy_obj

    @property
    def origin(self) -> scanAbstract:
        """
        Property for the original scan object.

        Returns
        -------
        scan_base
            The original scan object.
        """
        return self._origin

    @property
    def apply_to(self) -> list[str | int] | None:
        """
        The list of y-channel labels to apply the normalisation to.

        Parameters
        ----------
        apply_to : list[str | int] | None
            The new list of y-channel labels (or indexes) to apply the normalisation to.
            If None, applies to all channels.

        Returns
        -------
        list[str | int] | None
            The list of y-channel labels (or indexes) to apply the normalisation to.
            If None, applies to all channels.

        See Also
        --------
        scan_normalised.apply_to_indexes
            Method to return the indexes of the y-channels to apply the normalisation to.
        """
        return self._apply_to

    @apply_to.setter
    def apply_to(self, apply_to: list[str | int] | None) -> None:
        self._apply_to = apply_to

    @overload
    def apply_to_indexes(
        self, with_labels: Literal[True]
    ) -> tuple[list[int], list[str]]: ...

    @overload
    def apply_to_indexes(self, with_labels: Literal[False]) -> list[int]: ...

    @overload
    def apply_to_indexes(self) -> list[int]: ...

    def apply_to_indexes(
        self, with_labels: bool = False
    ) -> list[int] | tuple[list[int], list[str]]:
        """
        The list of y-channel indexes to apply the normalisation to.

        Any labels not found in the y_labels are ignored, and a warning is raised.

        Parameters
        ----------
        with_labels : bool, optional
            If True, returns a tuple of a list of indexes and a list of labels, by default False.

        Returns
        -------
        list[int] | tuple[list[int], list[str | None]]
            If `with_labels` is False, returns a list of indexes.
            If `with_labels` is True, returns a tuple of indexes and labels.
            If no `apply_to` is defined, returns a list of all indexes.
        """
        at = self._apply_to
        if at is not None:
            indexes: list[int] = []
            labels: list[str | None] = []
            ylabels = self.y_labels

            if ylabels is None:
                if any(isinstance(i, str) for i in at):
                    raise ValueError(
                        "Y labels are not defined in the scan object. Cannot match string labels in `apply_to`."
                    )

            # TODO: Add a check that indexes are within the y data range of the scan data / labels.

            # TODO: Consider parser relabels when looking for channel names.
            # Collect the relabels
            # parser = self.parser
            # relabels = parser.RELABELS

            for label in at:
                try:
                    indexes.append(ylabels.index(label))
                    labels.append(label)
                except ValueError:
                    warnings.warn(f"Label {label} not found in y_labels. ")
            if with_labels:
                return indexes, labels
            return indexes
        else:
            return list(range(len(self.y_labels)))

    @override
    def reload_labels_from_parser(self) -> None:
        return self._origin.reload_labels_from_parser()

    @property
    def ctime(self) -> datetime.datetime | None:
        """
        The creation time of the file as a datetime object.

        Returns
        -------
        datetime.datetime
            The creation time of the linked origin scan object.
        """
        origin = self._origin
        return origin.ctime if origin is not None else None

    @property
    def mtime(self) -> datetime.datetime | None:
        """
        The modification time of the file as a datetime object.

        Returns
        -------
        datetime.datetime
            The modification time of the linked origin scan object.
        """
        origin = self._origin
        return origin.mtime if origin is not None else None

    @property
    def filename(self) -> str | None:
        """
        Property for the filename of the scan object.

        Returns
        -------
        str | None
            The filename of the linked origin scan object.
        """
        origin = self._origin
        return origin.filename if isinstance(origin, scanBase) else None

    @abc.abstractmethod
    def _apply_normalisation(self) -> None:
        """
        Abstract method to apply the normalisation to the y data.
        """
        pass

    def _load_from_origin(self) -> None:
        """
        Method to reload data from the origin scan object.

        Uses instantiated scan reference.
        """
        # Copy data
        self._x = self._origin.x.copy() if self._origin.x is not None else None
        self._x_errs = (
            self._origin.x_errs.copy() if self._origin.x_errs is not None else None
        )
        self._y = self._origin.y.copy() if self._origin.y is not None else None
        self._y_errs = (
            self._origin.y_errs.copy() if self._origin.y_errs is not None else None
        )

    def load_and_normalise(self) -> None:
        """
        (Re)Load data from the origin and apply normalisation.

        Calls the `self._load_from_origin` method, which copies data from an original `scan` object,
        and then calls the `self._apply_normalisation` method, required to be implemented in subclasses.
        """
        self._load_from_origin()
        self._apply_normalisation()
        return

    ## X property already covered by scan_abstract

    @scanAbstract.x_label.getter
    def x_label(self) -> str:
        """
        Property for the x-axis label, from the origin scan object.

        Returns
        -------
        str
            The x-axis label.
        """
        return self.origin.x_label

    @x_label.setter
    def x_label(self, label: str | None) -> None:
        """
        Property setter for the x-axis label, to the origin scan.

        Parameters
        ----------
        label : str
            The new x-axis label.
        """
        self.origin.x_label = label

    @scanAbstract.x_unit.getter
    def x_unit(self) -> str:
        """
        Property for the x-axis unit, from the origin scan object.

        Returns
        -------
        str
            The x-axis unit.
        """
        return self.origin.x_unit

    @x_unit.setter
    def x_unit(self, unit: str) -> None:
        """
        Property setter for the x-axis unit, to the origin scan.

        Parameters
        ----------
        unit : str
            The new x-axis unit.
        """
        self.origin.x_unit = unit

    @scanAbstract.y.setter
    def y(
        self, vals: npt.NDArray | list[list[int | float]] | list[int | float]
    ) -> None:
        """
        Property setter for the y data.

        Similar to @scan_abstract.y, but avoids resetting the `y_units`
        and `y_labels` attributes when setting the y data dimension, as
        labels should always be inherited from the original scan object.

        Parameters
        ----------
        vals : npt.NDArray
            The new y data.
        """
        # Set Y values
        if isinstance(vals, np.ndarray):
            if len(vals.shape) == 2:
                self._y = vals.copy()
            else:
                # Add additional axis to make y 2D instead of 1D
                self._y = vals.copy()[:, np.newaxis]
        elif isinstance(vals, list):
            if isinstance(vals[0], list):
                self._y = np.array(vals)
            else:
                # Assuming int/floats, add additional axis as well.
                self._y = np.array(vals)[:, np.newaxis]
        else:
            raise ValueError("Is not a list or a np.ndarray.")

        # Remove errs if number of yvals have changed.
        if self.y_errs is not None and self.y_errs.shape != self._y.shape:
            self.y_errs = None

        # Do not remove unit/labels if number of yvals have changed!

    @scanAbstract.y_labels.getter
    def y_labels(self) -> list[str]:
        """
        Property for the y-axis labels, from the origin scan object.

        Parameters
        ----------
        labels : list[str] | None
            The new y-axis labels, propagated to the origin scan object.

        Returns
        -------
        list[str]
            The y-axis labels.
        """
        return self.origin.y_labels

    @y_labels.setter
    def y_labels(self, labels: list[str] | None) -> None:
        self.origin.y_labels = labels

    @scanAbstract.y_units.getter
    def y_units(self) -> list[str]:
        """
        Property for the y-axis units, from the origin scan object.

        Parameters
        ----------
        units : list[str]
            The new y-axis units, propagated to the origin scan object.

        Returns
        -------
        list[str]
            The y-axis units.
        """
        return self.origin.y_units

    @y_units.setter
    def y_units(self, units: list[str] | None) -> None:
        self.origin.y_units = units

    @abc.abstractmethod
    def _config_class(self) -> type[configBase]:
        """
        The settings class for the normalisation configuration.

        Allows loading from a settings object.

        Returns
        -------
        type[configBase]
            The configuration class for the normalisation settings.
        """
        pass

    @property
    @abc.abstractmethod
    def settings(self) -> configBase:
        """
        The accumulated normalisation settings.

        Returns
        -------
        norm_config
            The normalisation settings.
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def from_config(scan: scanAbstract, config: configBase) -> "scanAbstractNorm":
        """
        Load the normalisation settings from a configuration object.

        Parameters
        ----------
        scan : scan_abstract
            The original scan object to apply the normalisation to.
        config : normConfig
            The normalisation configuration object.

        Returns
        -------
        scanAbstractNorm
            A new instance of a scanAbstractNorm subclass initialised with the provided configuration.
        """
        classes = scanAbstractNorm.__subclasses__()
        for cls in classes:
            if config.__class__ == cls._config_class:
                if cls.from_config != scanAbstractNorm.from_config:
                    new_obj = cls.from_config(scan, config)
                    return new_obj
                else:
                    raise NotImplementedError(
                        f"Class {cls} has not implemented the `from_config` method."
                    )
        raise TypeError(
            f"No matching class found for the configuration class {config.__class__}."
        )


class scanNorm(scanAbstractNorm):
    """
    General class for a normalisation of a scanAbstract.

    Normalisation is usually performed in reference to some flux measurement,
    such as the current of a mesh or photodiode. This class requires
    `norm_channel` or `norm_data` to be defined, but not both.

    Parameters
    ----------
    scan : Type[scan_base]
        The initial scan object to collect and normalise y dataseries.
    norm_channel : str | int | None, optional
        The y-channel name corresponding to normalisation data.
        Can provide a string name or an integer index, but will be stored as a string.
    norm_method : normMethod, optional
        The normalisation method to apply to the normalisation channel. See `configChannel.normMethod` for options.
    apply_to : list[int] | list[str] | None, optional
        The list of y-channel indexes to apply the normalisation to.
    conv_indexes : bool, optional
        Whether to convert integer indexes in `apply_to` to string labels if available.

    Raises
    ------
    ValueError
        If both `norm_channel` and `norm_data` are simultaneously `None` or not-`None`. Only one can be defined.
    ValueError
        If `norm_channel` is not found in y_labels.

    See Also
    --------
    norm_settings.configChannel.normMethod
        The normalisation method enumerate used for the normalisation channel.
    """

    normMethod = configChannel.normMethod

    @override
    def __init__(
        self,
        scan: scanAbstract,
        norm_channel: str | int,
        norm_method: normMethod = configChannel.normMethod.DIV,
        apply_to: list[str | int] | None = None,
        conv_indexes: bool = True,
    ) -> None:
        # Set reference for original scan object
        super().__init__(scan, apply_to, conv_indexes=conv_indexes)

        self._norm_idx: int
        """The index of the normalisation channel in the origin scan"""

        # Store initial information.
        ylabels = self._origin.y_labels
        if ylabels is None:
            raise ValueError(
                "Y labels are not defined in the scan object. Cannot match string label for `norm_channel`."
            )
        if isinstance(norm_channel, str):
            self._norm_idx = ylabels.index(norm_channel)
        else:
            self._norm_idx = norm_channel
            ylabel = ylabels[norm_channel]
            if ylabel is None:
                raise ValueError(
                    f"Y label at the specified index `{norm_channel}` is not defined. Cannot match string label for `norm_channel`."
                )
            norm_channel = ylabel

        # Save the normalisation channel settings as an object.
        self._config = configChannel(
            norm_method=norm_method, channel_name=norm_channel, apply_to=apply_to
        )
        """The normalisation configuration object used for serialising settings."""

        # Process normalisation data
        self.load_and_normalise()

    @override
    def _config_class(self) -> type[configBase]:
        """
        Return the settings class for the normalisation configuration.

        Required to allows loading from a settings object.

        Returns
        -------
        type[configChannel]
            The configuration class for the normalisation settings.
        """
        return configChannel

    @scanAbstractNorm.settings.getter
    @override
    def settings(self) -> configChannel:
        """
        A settings object capturing the settings for the normalisation channel.

        Returns
        -------
        configChannel
            The normalisation settings object, containing the channel and method.
        """
        return self._config

    @override
    @staticmethod
    def from_config(scan: scanAbstract, config: configChannel) -> scanNorm:  # type: ignore # TODO: implement fix for subtype override.
        """
        Load the normalisation settings from a configuration object.

        Parameters
        ----------
        scan : scan_abstract
            The original scan object to apply the normalisation to.
        config : configChannel
            The normalisation configuration object.

        Returns
        -------
        scanNorm
            A new instance of scanNorm initialised with the provided configuration.
        """
        return scanNorm(
            scan=scan, norm_channel=config.channel, norm_method=config.method
        )

    @property
    def method(self) -> normMethod:
        """
        The normalisation method used for the normalisation channel.

        Parameters
        ----------
        method : normMethod
            The new normalisation method enumerate. Can be
            - 0: `normMethod.NONE` for no normalisation.
            - 1: `normMethod.SUB` for background subtraction.
            - 2: `normMethod.DIV` for normalisation by amplitude variation.

        Returns
        -------
        normMethod
            The normalisation method.
        """
        return self._config.method

    @method.setter
    def method(self, method: normMethod) -> None:
        self._config.method = method

    @property
    def channel(self) -> str | None:
        """
        The normalisation channel name.

        Parameters
        ----------
        channel : str
            The new normalisation channel name.

        Returns
        -------
        str
            The normalisation channel name.
        """
        return self._config.channel

    @channel.setter
    def channel(self, channel: str) -> None:
        self._config.channel = channel

    @override
    def _load_from_origin(self) -> None:
        """
        Re-load from source, refreshing data, errors, labels and units for x,y variables.

        Overrides `scan_abstract_norm._load_from_origin` depending on the `_norm_channel` attribute being None.
        """
        # Y Reloading
        y_labels = self._origin.y_labels
        channel = self.channel
        ind = self._norm_idx
        if channel is not None:
            # Collect index of normalisation data
            if y_labels is not None:
                self._norm_idx = ind = self._origin.y_labels.index(channel)
            else:
                raise ValueError(
                    "Y labels are not defined in the scan object. Cannot match string label for `norm_channel`."
                )
        else:
            # idx is not None.
            if y_labels is not None:
                label = y_labels[ind]
                if label is not None:
                    self.channel = label

        # Collect normalisaton data:
        self._norm_data = self._origin.y[:, ind]
        self._norm_data_errs = (
            self._origin.y_errs[:, ind] if self._origin.y_errs is not None else None
        )

        ## Old method to remove normalisation channel from y data.
        ## We don't do this anymore, as it causes confusion when chaining normalisations.
        ## For example, if we're preserving the indexes of channels to apply to.

        # # Collect Y data, removing index of normalisation channel.
        # self._y = np.delete(
        #     self._origin.y, ind, 1
        # )  # copy data, removing the index from the existing set of data.
        # self._y_errs = (
        #     np.delete(self._origin.y_errs, ind, 1)
        #     if self._origin.y_errs is not None
        #     else None
        # )

        # # Collect X data normally.
        # self._x = self._origin.x.copy()
        # self._x_errs = (
        #     self._origin.x_errs.copy() if self._origin.x_errs is not None else None
        # )

        # Load x and y data regularly.
        super()._load_from_origin()

    @override
    def _apply_normalisation(self) -> None:
        """
        Perform the selected normalisation.

        `self.method` determines the method to apply to the provided y normalisation data.
        """
        match self.method:
            case self.normMethod.NONE:
                # Do nothing for no normalisation
                pass
            case self.normMethod.SUB:
                # Subtract the normalisation data from the y data.
                self._normalisation_subtract()
            case self.normMethod.DIV:
                # Scale the y data by the normalisation data.
                self._normalisation_divide()
            case self.normMethod.MULT:
                self._normalisation_multiply()
            case _:
                raise ValueError("Normalisation method not defined.")
        return

    def _normalisation_multiply(self) -> None:
        # Multiply the y data by the normalisation data.
        ylabels = self.y_labels
        at = self._apply_to
        if (
            ylabels is None
            and at is not None
            and not all(isinstance(i, int) for i in at)
        ):
            raise ValueError(
                "Y labels are not defined in the scan object. Cannot match string labels in `apply_to`."
            )
        y = self._y
        if y is None:
            raise ValueError("Y data is not defined.")

        for i in range(y.shape[1]):
            label = ylabels[i] if ylabels is not None else None
            if i == self._norm_idx:
                # Skip normalisation channel
                continue
            if at is None or (label in at or i in at):
                y[:, i] *= self._norm_data

    def _normalisation_divide(self) -> None:
        """
        Scale the y data by the normalisation data.

        Applies the amplitude normalisation to `apply_to` channels, otherwise to all channels.
        """
        # Scale the y data, only in the apply_to channels.
        ylabels = self.y_labels
        at = self._apply_to
        if (
            ylabels is None
            and at is not None
            and not all(isinstance(i, int) for i in at)
        ):
            raise ValueError(
                "Y labels are not defined in the scan object. Cannot match string labels in `apply_to`."
            )
        y = self._y
        if y is None:
            raise ValueError("Y data is not defined.")

        for i in range(y.shape[1]):
            label = ylabels[i] if ylabels is not None else None
            if i == self._norm_idx:
                # Skip normalisation channel
                continue
            if at is None or (label in at or i in at):
                y[:, i] /= self._norm_data

    def _normalisation_subtract(self) -> None:
        """
        Subtract the background from the selected y data.

        Applies the background subtraction to `apply_to` channels, otherwise to all channels.
        """
        # Subtract the normalisation data from the y data.
        ylabels = self.y_labels
        at = self._apply_to
        if (
            ylabels is None
            and at is not None
            and not all(isinstance(i, int) for i in at)
        ):
            raise ValueError(
                "Y labels are not defined in the scan object. Cannot match string labels in `apply_to`."
            )
        y = self._y
        if y is None:
            raise ValueError("Y data is not defined.")

        for i in range(y.shape[1]):
            label = ylabels[i] if ylabels is not None else None
            if i == self._norm_idx:
                # Skip normalisation channel
                continue
            if at is None or (label in at or i in at):
                y[:, i] -= self._norm_data

    @scanAbstractNorm.y_labels.getter
    def y_labels(self) -> list[str | None] | None:
        """
        Property for the y-axis labels, from the origin scan object.

        Parameters
        ----------
        labels : list[str] | None
            The new y-axis labels.

        Returns
        -------
        list[str]
            The y-axis labels.
        """
        origin_labels = self._origin.y_labels
        ## No longer removing the normalisation channel from the labels, as it causes confusion when chaining normalisations.
        # idx = self._norm_idx
        # if idx is not None:
        #     return origin_labels[0:idx] + origin_labels[idx + 1 :]
        # else:
        return origin_labels

    @y_labels.setter
    def y_labels(self, labels: list[str] | None) -> None:
        idx = self._norm_idx
        if idx is not None and labels is not None and self.channel is not None:
            # Include the normalisation channel in the labels when setting.
            self._origin.y_labels = labels[0:idx] + [self.channel] + labels[idx:]
        else:
            # Otherwise set labels normally.
            self._origin.y_labels = labels

    @scanAbstractNorm.y_units.getter
    def y_units(self) -> list[str] | None:
        """
        Property for the y-axis units, from the origin scan object.

        When the normalisation channel is defined, the units are adjusted to remove the normalisation channel.

        Parameters
        ----------
        labels : list[str] | None
            The new y-axis units.

        Returns
        -------
        list[str] | None
            The y-axis units.
        """
        origin_units = self._origin.y_units
        idx = self._norm_idx
        if idx is not None and origin_units is not None:
            return origin_units[0:idx] + origin_units[idx + 1 :]
        else:
            return origin_units

    @y_units.setter
    def y_units(self, units: list[str] | None) -> None:
        idx = self._norm_idx
        if idx is not None:
            if units is not None:
                # Include the normalisation channel in the units when setting.
                self._origin.y_units = units[0:idx] + [self.channel] + units[idx:]
                return
        # Otherwise, set units normally.
        self._origin.y_units = units
        return

    def copy(self) -> scanNorm:
        """
        A copy of the scan object.

        Returns
        -------
        scan_abstract
            A copy of the scan object.
        """
        return scanNorm(
            scan=self._origin,
            norm_channel=self.channel,
            norm_method=self.method,
            apply_to=self._apply_to,
        )


class scanNormExt(scanNorm):
    """
    Normalising a scan_base to a particular channel of a background/external scan.

    Useful for NEXAFS such as carbon K edges. If a mesh current is used to normalise the signal,
    it will not necessarily normalise to the true absorption signal. A downstream photodiode, or alternative
    intensity channel, can be used to capture a true intensity signal. This channel can then
    permit double normalisation (i.e., with the mesh current signal) on the measured spectra.

    Parameters
    ----------
    scan : scanAbstract
        The initial scan object to collect and normalise.
    ext_scan : scanAbstract | str
        The background scan object to use for normalisation.
        Can also be a string filename or folder path.
    norm_channel : str
        The y-channel name corresponding to normalisation data in the background scan.
        This channel will be used to normalise the y data of the scan object.
    norm_method : normMethod, optional
        The normalisation method to apply to the normalisation channel. See `configChannel.normMethod` for options.
    apply_to : list[int] | list[str] | None, optional
        The list of y-channel labels to apply the normalisation to.
    conv_indexes : bool, optional
        Whether to convert integer indexes in `apply_to` to string labels if available.

    See Also
    --------
    norm_settings.normMethod
        The normalisation method enumerate used for the normalisation channel.
    norm_settings.extSelection
        The method to select an external scan from a folder path.

    Notes
    -----
    This class extends `scanNorm` to use a different scan object for normalisation data, but uses the same methods
    for normalisation (subtraction, division, multiplication).
    """

    normMethod = normMethod
    extSelection = extSelection

    @override
    def __init__(
        self,
        scan: scanAbstract,
        ext_scan: scanAbstract,
        norm_channel: str | int,
        norm_method: normMethod = configChannel.normMethod.DIV,
        apply_to: list[str | int] | None = None,
        conv_indexes: bool = True,
    ) -> None:
        # Do not call super() for scanNorm, as we have a different configuration.
        # Set reference for original scan object
        scanAbstractNorm.__init__(self, scan, apply_to, conv_indexes=conv_indexes)

        # New variables
        self._norm_idx: int
        """The index of the normalisation channel in the background scan"""
        self._ext_scan: scanAbstract = ext_scan
        """The external/background scan object used for normalisation."""

        # Store initial information.
        ext_y_labels = ext_scan.y_labels
        if ext_y_labels is not None:
            if isinstance(norm_channel, str):
                self._norm_idx = ext_y_labels.index(norm_channel)
                channel = norm_channel
            else:
                self._norm_idx = norm_channel
                channel = ext_y_labels[norm_channel]
        else:
            raise ValueError(
                f"External scan has no y_labels defined. Cannot match normalisation channel name `{norm_channel}`."
            )

        # Redefine the normalisation index
        ext_ylabels = ext_scan.y_labels
        if ext_ylabels is None:
            raise ValueError(
                "Y labels are not defined in the external scan object. Cannot match string label for `norm_channel`."
            )
        self._norm_idx: int = (
            ext_ylabels.index(norm_channel)
            if isinstance(norm_channel, str)
            else norm_channel
        )

        # Save the normalisation channel settings as an object.
        self._config = configExternalChannel(
            channel_selection=extSelection.FIXED_SCAN,
            path=ext_scan.filepath,
            keyword=None,
            norm_method=norm_method,
            channel_name=channel,
            apply_to=apply_to,
        )
        """The normalisation configuration object used for serialising settings."""

        ### TODO: Make extra parameters into a configuration object that deals with folders rather than single files.
        # """
        # ext_selection : extSelection, optional
        #     The method to select the external scan when providing a folder path.
        #     Options are defined in `extSelection` enumerate.
        #     Defaults to `extSelection.FIXED_SCAN`, when a single filename is provided.
        # ext_keyword : str | None, optional
        #     The keyword to filter by, when selecting an external scan from a folder.
        # """
        # # If ext_scan is a string
        # path: str | None = None
        # if isinstance(ext_scan, str):
        #     if os.path.isfile(ext_scan):
        #         assert extSelection == extSelection.FIXED_SCAN, "When providing a single filename for ext_scan, ext_selection must be FIXED_SCAN."
        #         path = ext_scan
        #     elif os.path.isdir(ext_scan):
        #         assert extSelection != extSelection.FIXED_SCAN, "When providing a folder path for ext_scan, ext_selection cannot be FIXED_SCAN."
        #         path = ext_scan
        #     else:
        #         raise ValueError(f"Provided ext_scan string `{ext_scan}` is not a valid file or folder path.")
        # elif isinstance(ext_scan, scanAbstract):
        #     # Get the filename from the scan object
        #     path = ext_scan.filepath
        # else:
        #     raise TypeError(f"`ext_scan` must be a scanAbstract object or a string file/folder path. Got {type(ext_scan)}.")

        # Process normalisation data
        self.load_and_normalise()

    @override
    def _config_class(self) -> type[configBase]:
        """
        Return the settings class for the normalisation configuration.

        Required to allows loading from a settings object.

        Returns
        -------
        type[configChannel]
            The configuration class for the normalisation settings.
        """
        return configExternalChannel

    @scanNorm.channel.getter
    def channel(self) -> str | None:
        """
        The normalisation channel name from the background scan.

        Parameters
        ----------
        channel : str | None
            The new normalisation channel name.

        Returns
        -------
        str | None
            The normalisation channel name.
        """
        return self._config.channel

    @channel.setter
    def channel(self, channel: str | None) -> None:
        self._config.channel = channel

    @channel.deleter
    def channel(self) -> None:
        self._config.channel = None

    @override
    def _load_from_origin(self) -> None:
        """
        Re-load from source, refreshing data, errors, labels and units for x,y variables.

        Overrides `scanNorm._load_from_origin` depending on the `_norm_channel` attribute being None.
        """
        # Y Reloading
        channel = self.channel
        ext_scan = self._ext_scan
        if ext_scan is not None:
            ext_ylabels = ext_scan.y_labels
            # Update using the channel name
            if ext_ylabels is not None:
                if channel is not None:
                    self._norm_idx = ind = ext_ylabels.index(channel)
                    self.channel = channel
                else:
                    self.channel = ext_ylabels[self._norm_idx]
                    ind = self._norm_idx
            else:
                raise ValueError("External scan y_labels are not defined.")

            # Collect normalisaton data:
            ext_y = ext_scan.y
            if ext_y is None:
                raise ValueError("External scan y data is not defined.")
            self._norm_data = ext_y[:, ind].copy() if ext_y is not None else None
            self._norm_data_errs = (
                self._ext_scan.y_errs[:, ind].copy()
                if self._ext_scan.y_errs is not None
                else None
            )
        else:
            raise ValueError("External scan is not defined.")

        # Load x and y data regularly; skip the `scanNorm._load_from_origin` to avoid using `self._norm_idx`
        scanAbstractNorm._load_from_origin(self)


class scanNormEdges(scanAbstractNorm):
    """
    Normalising a scan_base across pre &/ post edges.

    Uses two band definitions to normalise the y data centred around a chemical resonant edge
    such as the K-edge of carbon.

    Parameters
    ----------
    scan : Type[scan_base] | scan_base
        The initial scan object to collect and normalise y dataseries.
    pre_edge_domain : _type_, optional
        Data to define the domain of pre-edge normalisation. Can take the following data forms:
        - A list of indices (integers) to subset the y data.
        - A tuple of floats to define the inclusive-domain of the x data.
        - None, to not perform pre-edge normalisation. Overrides pre_edge_normalisation enumerate.
    post_edge_domain : _type_, optional
        Data to define the domain of post-edge normalisation. Same format as pre-edge.
        If None overrides post_edge_normalisation enumerate.
    pre_edge_norm_method : edgeNormPre, optional
        Normalisation type for pre-edge, by default edgeNormPre.LINEAR.
    post_edge_norm_method : EDGE_NORM_TYPE, optional
        Normalisation type for post-edge, by default EDGE_NORM_TYPE.LINEAR.
    pre_edge_level : float, optional
        Normalisation level for pre-edge, by default 0.1.
    post_edge_level : float, optional
        Normalisation level for post-edge, by default 1.0.
    apply_to : list[str] | None, optional
        The list of y-channel labels to apply the normalisation to.

    Attributes
    ----------
    edgeNormPre : enumerate
        Enumerate types for pre edge normalisation.
    edgeNormPost : enumerate
        Enumerate types for post edge normalisation.

    Raises
    ------
    ValueError
        If both `norm_channel` and `norm_data` are simultaneously None or not-None. Only one can be defined.
    ValueError
        If `norm_channel` is not found in y_labels.
    """

    @staticmethod
    def LIN_FN(x: npt.NDArray, m: float, c: float) -> npt.NDArray:
        return m * x + c

    """
    A function to define a linear fit.

    Parameters
    ----------
    x : ArrayLike
        The x value.
    m : float
        The gradient of the line.
    c : float
        The y-intercept of the line.
    """

    @staticmethod
    def EXP_FN_OFFSET(x: npt.NDArray, a: float, b: float, c: float) -> npt.NDArray:
        return a * np.exp(b * (x - x.min())) + c

    """
    A function to define an exponential fit, adjusted to the minimum x value.

    Parameters
    ----------
    x : ArrayLike
        The x value.
    a : float
        The amplitude of the exponential.
    b : float
        The decay constant of the exponential.
    c : float
        The offset of the exponential.
    """

    edgeNormPre = edgeNormPre
    edgeNormPost = edgeNormPost

    def __init__(
        self,
        scan: scanAbstract,
        pre_edge_domain: tuple[float, float] | list[int] | None,
        post_edge_domain: tuple[float, float] | list[int] | None,
        pre_edge_norm_method: edgeNormPre = edgeNormPre.CONSTANT,
        post_edge_norm_method: edgeNormPost = edgeNormPost.CONSTANT,
        pre_edge_level: float | None = None,
        post_edge_level: float | None = None,
        apply_to: list[str] | None = None,
    ) -> None:
        # Initialize the scan_abstract_norm class.
        super().__init__(scan=scan, apply_to=apply_to)
        # Check level defaults
        if pre_edge_level is None:
            pre_edge_level = configEdges.DEFAULT_PRE_EDGE_LEVEL_CONSTANT
        if post_edge_level is None:
            post_edge_level = configEdges.DEFAULT_POST_EDGE_LEVEL_CONSTANT
        # Create a settings object to store the normalisation settings.
        self._config = configEdges(
            pre_edge_domain=pre_edge_domain,
            post_edge_domain=post_edge_domain,
            pre_edge_norm_method=pre_edge_norm_method,
            post_edge_norm_method=post_edge_norm_method,
            pre_edge_level=pre_edge_level,
            post_edge_level=post_edge_level,
            apply_to=apply_to,
        )

        # Define variables for edge fitting
        self.pre_edge_fit_params: list | None = None
        self.post_edge_fit_params: list | None = None

        # Perform normalisation
        self.load_and_normalise()
        return

    @override
    def _config_class(self) -> type[configBase]:
        """
        The linked configuration class for the normalisation settings.

        Returns
        -------
        type[configEdges]
            The configuration class for the normalisation settings.
        """
        return configEdges

    @scanAbstractNorm.settings.getter
    @override
    def settings(self) -> configEdges:
        """
        A settings object associated with the settings for a normalisation channel.

        Returns
        -------
        configEdges
            The normalisation settings.
        """
        return self._config

    @override
    @staticmethod
    def from_config(scan: scanAbstract, config: configEdges) -> scanNormEdges:  # type: ignore # TODO: implement fix for subtype override.
        """
        Load the normalisation settings from a configuration object.

        Parameters
        ----------
        scan : scan_abstract
            The original scan object to apply the normalisation to.
        config : configEdges
            The normalisation configuration object.

        Returns
        -------
        scanNormEdges
            A new instance of scanNormEdges with the provided configuration.
        """
        return scanNormEdges(
            scan=scan,
            pre_edge_domain=config.pre_edge_domain,
            post_edge_domain=config.post_edge_domain,
            pre_edge_norm_method=config.pre_edge_norm_method,
            post_edge_norm_method=config.post_edge_norm_method,
            pre_edge_level=config.pre_edge_level,
            post_edge_level=config.post_edge_level,
            apply_to=config.apply_to,
        )

    @override
    def copy(self) -> scanNormEdges:
        clone = scanNormEdges(
            self._origin,
            pre_edge_domain=self._config._pre_edge_domain,
            post_edge_domain=self._config._post_edge_domain,
            pre_edge_norm_method=self._config._pre_edge_norm_method,
            post_edge_norm_method=self._config._post_edge_norm_method,
            pre_edge_level=self._config._pre_edge_level,
            post_edge_level=self._config._post_edge_level,
            apply_to=self._config._apply_to,
        )
        return clone

    @property
    def DEFAULT_PRE_EDGE_LEVEL_CONSTANT(self) -> float:
        return self._config.DEFAULT_PRE_EDGE_LEVEL_CONSTANT

    @property
    def DEFAULT_PRE_EDGE_LEVEL_LINEAR(self) -> float:
        return self._config.DEFAULT_PRE_EDGE_LEVEL_LINEAR

    @property
    def DEFAULT_PRE_EDGE_LEVEL_EXP(self) -> float:
        return self._config.DEFAULT_PRE_EDGE_LEVEL_EXP

    @property
    def DEFAULT_POST_EDGE_LEVEL_CONSTANT(self) -> float:
        return self._config.DEFAULT_POST_EDGE_LEVEL_CONSTANT

    @override
    def _apply_normalisation(self) -> None:
        """
        Scale y data by the normalisation regions.

        Applies the pre-edge and post-edge normalisation methods to the y data.
        """
        # Define fitting functions
        lin_fn = self.LIN_FN
        exp_fn = self.EXP_FN_OFFSET
        # Perform dual normalisation, sequentially to avoid duplicate code.

        ## PRE-EDGE
        if (
            self.pre_edge_norm_method is not edgeNormPre.NONE
            and self.pre_edge_domain is not None
        ):
            # Collect pre-edge indexes
            if isinstance(self.pre_edge_domain, list):
                # Use the list of indexes
                pre_inds = self.pre_edge_domain
                if len(pre_inds) == 0:
                    raise ValueError("Pre-edge index list is empty.")
            elif isinstance(self.pre_edge_domain, tuple):
                # Use the domain range
                pre_inds = np.asarray(
                    (self.x >= self.pre_edge_domain[0])
                    & (self.x <= self.pre_edge_domain[1])
                ).nonzero()
                # Check indexes of each tuple element
                if len(pre_inds[0]) == 0:
                    raise ValueError(
                        f"Pre-edge domain ({self.pre_edge_domain[0]} to {self.pre_edge_domain[1]}) contains no datapoints."
                    )
            else:
                raise AttributeError(
                    "Pre-edge domain is not defined correctly. Should be a list of indexes or the range in a tuple."
                )

            # Calculate pre-edge on all y values and normalise
            match self.pre_edge_norm_method:
                case edgeNormPre.CONSTANT:
                    if self.apply_to:
                        inds = self.apply_to_indexes()
                        mean = np.mean(self.y[pre_inds][:, inds], axis=0)
                        stdev = np.std(self.y[pre_inds][:, inds], axis=0)
                        self.y[:, inds] += -mean[np.newaxis, :] + self.pre_edge_level
                        self.pre_edge_fit_params = mean.tolist()

                        if self.y_errs is not None:
                            # Add the standard deviation
                            self.y_errs[:, inds] = np.sqrt(
                                np.square(self.y_errs[:, inds])
                                + stdev[np.newaxis, :] ** 2
                            )

                    else:
                        mean = np.mean(self.y[pre_inds], axis=0)
                        stdev = np.std(self.y[pre_inds], axis=0)

                        self.y[:, :] += -mean[np.newaxis, :] + self.pre_edge_level
                        self.pre_edge_fit_params = mean.tolist()

                        if self.y_errs is not None:
                            # Add the standard deviation
                            self.y_errs = np.sqrt(
                                np.square(self.y_errs) + stdev[np.newaxis, :] ** 2
                            )

                case edgeNormPre.LINEAR:
                    # TODO: Implement apply to
                    params: list = []
                    for i in range(self.y.shape[1]):
                        popt, pcov = sopt.curve_fit(
                            lin_fn, self.x[pre_inds], self.y[:, i][pre_inds]
                        )
                        self.pre_edge_fit_params = popt.tolist()
                        self.y[:, i] += -lin_fn(self.x, *popt) + self.pre_edge_level
                        params += popt.tolist()

                    self.pre_edge_fit_params = params

                case edgeNormPre.EXPONENTIAL:
                    # TODO: Implement apply to
                    params: list = []
                    for i in range(self.y.shape[1]):
                        # Subtract offset to fit
                        xfit = self.x[pre_inds] - self.x[pre_inds].min()
                        x = self.x - self.x[pre_inds].min()
                        a0 = self.y[:, i][pre_inds].max() - self.y[:, i][pre_inds].min()
                        # Fit
                        popt, pcov = sopt.curve_fit(
                            exp_fn,
                            xfit,
                            self.y[:, i][pre_inds],
                            p0=(a0, -0.5, 0),
                            bounds=[(-np.inf, -np.inf, -np.inf), (np.inf, 0, np.inf)],
                            maxfev=10000,
                        )
                        self.y[:, i] += -exp_fn(x, *popt) + self.pre_edge_level
                        params += popt.tolist()

                    self.pre_edge_fit_params = params

                case edgeNormPre.NONE:
                    # Do nothing for no pre-edge normalisation
                    pass
                case _:
                    # Should never reach here, and dual normalisation excludes None type.
                    raise ValueError("Pre-edge normalisation type not defined.")

        ## POST-EDGE
        if (
            self.post_edge_norm_method is not edgeNormPost.NONE
            and self.post_edge_domain is not None
        ):
            # Collect post-edge indexes:
            if isinstance(self.post_edge_domain, list):
                post_inds = self.post_edge_domain
                if len(post_inds) == 0:
                    raise ValueError("Post-edge index list is empty.")
            elif isinstance(self.post_edge_domain, tuple):
                post_inds = np.where(
                    (self.x >= self.post_edge_domain[0])
                    & (self.x <= self.post_edge_domain[1])
                )
                # Check indexes of each tuple element
                if len(post_inds[0]) == 0:
                    raise ValueError(
                        f"Post-edge domain ({self.post_edge_domain[0]} to {self.post_edge_domain[1]}) contains no datapoints."
                    )
            else:
                raise AttributeError(
                    "Post-edge domain is not defined correctly. Should be a list of indexes or the range in a tuple."
                )

            if len(post_inds) == 0:
                raise ValueError("Post-edge domain is empty.")

            # Calculate post-edge and normalise. Note, y data is already normalised by pre-edge.
            match self.post_edge_norm_method:
                case edgeNormPost.CONSTANT:
                    # TODO: Implement apply to
                    if self.apply_to:
                        inds = self.apply_to_indexes()
                        postave = np.mean(self.y[post_inds][:, inds], axis=0)
                        scale = np.zeros(postave.shape)
                        scale[postave != 0] = (
                            self.post_edge_level - self.pre_edge_level
                        ) / postave[postave != 0]
                        self.y[:, inds] = (
                            self.y[:, inds] - self.pre_edge_level
                        ) * scale[np.newaxis, :] + self.pre_edge_level
                    else:
                        # Get the mean of the post-edge from the pre-edge level.
                        postave = np.mean(
                            self.y[post_inds] - self.pre_edge_level, axis=0
                        )
                        # Scale to difference from pre-edge level to post-edge level. Ignore zero values.
                        scale = np.zeros(postave.shape)
                        scale[postave != 0] = (
                            self.post_edge_level - self.pre_edge_level
                        ) / postave[postave != 0]
                        self.y = (
                            self.y - self.pre_edge_level
                        ) * scale + self.pre_edge_level
                case edgeNormPost.NONE:
                    # Do nothing
                    pass
                case _:
                    # Should never reach here, and dual normalisation excludes None type.
                    raise ValueError(
                        f"Post-edge normalisation type not defined for {self.post_edge_norm_method}."
                    )

        return

    @property
    def pre_edge_domain(self) -> tuple[float, float] | list[int] | None:
        """
        A property defining the pre-edge domain of normalisation.

        If setting and `pre_edge_normalisation` is `NONE`, will set new normalisation enumerate to `LINEAR`.

        Parameters
        ----------
        vals : list[int] | tuple[float, float] | None
            Can be defined using a list of included indices matching x datapoints.
            Alternatively a tuple of two floats defining the inclusive endpoints.
            Alternatively None stops normalisation being performed on the pre-edge.

        Returns
        -------
        list[int] | tuple[float, float] | None
            Same as vals parameter.
        """
        return self._config._pre_edge_domain

    @pre_edge_domain.setter
    def pre_edge_domain(self, vals: list[int] | tuple[float, float] | None) -> None:
        self._config._pre_edge_domain = vals

    @pre_edge_domain.deleter
    def pre_edge_domain(self):
        del self._config._pre_edge_domain

    @property
    def pre_edge_norm_method(self) -> edgeNormPre:
        """
        Property to define the type of normalisation performed on the pre-edge.

        Parameters
        ----------
        vals : edgeNormPre
            LINEAR, EXPONENTIAL or NONE.

        Returns
        -------
        edgeNormPre
            The current normalisation type.
        """
        return self._config.pre_edge_norm_method

    @pre_edge_norm_method.setter
    def pre_edge_norm_method(self, vals: edgeNormPre) -> None:
        self._config.pre_edge_norm_method = vals

    @pre_edge_norm_method.deleter
    def pre_edge_norm_method(self) -> None:
        del self._config.pre_edge_norm_method

    @property
    def pre_edge_level(self) -> float:
        """
        Property to define the normalisation level (constant) for the pre-edge.

        Parameters
        ----------
        vals : float
            The normalisation average.

        Returns
        -------
        float
            The normalisation level.
        """
        return self._config.pre_edge_level

    @pre_edge_level.setter
    def pre_edge_level(self, level: float) -> None:
        self._config.pre_edge_level = level

    @pre_edge_level.deleter
    def pre_edge_level(self) -> None:
        del self._config.pre_edge_level

    @property
    def post_edge_domain(self) -> list[int] | tuple[float, float] | None:
        """
        A property defining the post-edge domain of normalisation.

        If setting and `post_edge_normalisation` is `NONE`, will set new normalisation enumerate to `LINEAR`.

        Parameters
        ----------
        vals : list[int] | tuple[float, float] | None
            Can be defined using a list of included indices matching x datapoints.
            Alternatively a tuple of two floats defining the inclusive endpoints.
            Alternatively None stops normalisation being performed on the post-edge.

        Returns
        -------
        list[int] | tuple[float, float] | None
            Same as vals parameter.
        """
        return self._config._post_edge_domain

    @post_edge_domain.setter
    def post_edge_domain(self, vals: list[int] | tuple[float, float] | None) -> None:
        self._config.post_edge_domain = vals

    @post_edge_domain.deleter
    def post_edge_domain(self):
        del self._config.post_edge_domain

    @property
    def post_edge_norm_method(self) -> edgeNormPost:
        """
        Property to define the type of normalisation performed on the post-edge.

        Parameters
        ----------
        vals : edgeNormPost
            LINEAR, EXPONENTIAL or NONE.

        Returns
        -------
        edgeNormPost
            The current normalisation type.
        """
        return self._config._post_edge_norm_method

    @post_edge_norm_method.setter
    def post_edge_norm_method(self, method: edgeNormPost) -> None:
        self._config.post_edge_norm_method = method

    @post_edge_norm_method.deleter
    def post_edge_norm_method(self) -> None:
        self._post_edge_normalisation = None

    @property
    def post_edge_level(self) -> float:
        """
        Property to define the normalisation level (constant) for the post-edge.

        Parameters
        ----------
        vals : float
            The normalisation average.

        Returns
        -------
        float
            The normalisation level.
        """
        return self._config.post_edge_level

    @post_edge_level.setter
    def post_edge_level(self, vals: float) -> None:
        # if (
        #     self.post_edge_normalisation
        #     is edgeNormPost.EXPONENTIAL
        # ):
        #     if vals <= 0:
        #         raise ValueError(
        #             "Exponential normalisation requires a positive, non-zero level."
        #         )
        self._config.post_edge_level = vals

    @post_edge_level.deleter
    def post_edge_level(self) -> None:
        if (
            # self.post_edge_normalisation
            # is edgeNormPost.EXPONENTIAL
            False
        ):
            self._post_edge_level = configEdges.DEFAULT_PRE_EDGE_LEVEL_EXP
        # elif self.pre_edge_normalisation is configEdges.EDGE_NORM_TYPE.LINEAR:
        # self._pre_edge_level = configEdges.DEFAULT_PRE_EDGE_LEVEL_LINEAR
        else:
            self._post_edge_level = configEdges.DEFAULT_PRE_EDGE_LEVEL_LINEAR


class scanDoubleNorm(scanAbstractNorm):
    r"""
    A scan object normalised by the combination of an internal and external channel.

    The normalisation is calculated by dividing the y data by the internal normalisation channel,
    then multiplying by the same normalisation channel from an external scan object, before dividing
    by a second normalisation channel from the external scan object (i.e. a direct photodiode).

    .. math::
        y_{norm} = \frac{y_{scan}}{y_{norm\_channel}} \times \frac{y_{ext\_norm\_channel}}{y_{ext\_double\_norm\_channel}}

    Parameters
    ----------
    scan : scanAbstract
        The initial scan object to collect and normalise.
    ext_scan : scanAbstract
        The external scan object to use for normalisation.
    norm_channel : str | int
        The y-channel name corresponding to normalisation data in the original scan.
    double_norm_channel : str | int
        The y-channel name corresponding to normalisation data in the external scan.
    apply_to : list[str | int] | None, optional
        The list of y-channel labels to apply the normalisation to.
    conv_indexes : bool, optional
        Whether to convert indexes in apply_to to labels.
    """

    def __init__(
        self,
        scan: scanAbstract,
        ext_scan: scanAbstract,
        norm_channel: str | int,
        double_norm_channel: str | int,
        apply_to: list[str | int] | None = None,
        conv_indexes: bool = True,
    ) -> None:
        ext_ylabels = ext_scan.y_labels
        ylabels = scan.y_labels
        if isinstance(norm_channel, int) or isinstance(double_norm_channel, int):
            assert ylabels == ext_ylabels, (
                "When providing integer indexes for both normalisation channels, the y_labels of both scans must \
                match to ensure normalisation is being applied correctly."
            )

        # Automatically remove the channels from the apply_to list.
        if isinstance(double_norm_channel, int):
            double_norm_channel_idx = double_norm_channel
        else:
            if ext_ylabels is None:
                raise ValueError("External scan y_labels are not defined.")
            else:
                double_norm_channel_idx = ext_ylabels.index(double_norm_channel)

        if isinstance(norm_channel, int):
            norm_channel_idx = norm_channel
        else:
            if ylabels is None:
                raise ValueError("Scan y_labels are not defined.")
            else:
                norm_channel_idx = ylabels.index(norm_channel)

        if apply_to is not None:
            apply_to = [
                a
                for a in apply_to
                # Avoid all four possibilities
                if a != norm_channel
                and a != double_norm_channel
                and a != norm_channel_idx
                and a != double_norm_channel_idx
            ]
        else:
            apply_to = [
                a
                for a in (list(range(len(ylabels))) if ylabels is not None else [])
                # Avoid both index possibilities
                if a != norm_channel_idx and a != double_norm_channel_idx
            ]

        # Initialize the scan_abstract_norm class.
        scanAbstractNorm.__init__(
            self, scan, apply_to=apply_to, conv_indexes=conv_indexes
        )

        # Save the reference to the norm scan
        self._ext_scan: scanAbstract = ext_scan
        """The external/background scan object used for normalisation."""
        self._orig_norm_channel: str | int = norm_channel
        """The original normalisation channel from the scan object."""
        self._orig_double_norm_channel: str | int = double_norm_channel
        """The original double normalisation channel from the external scan object."""

        # Use the existing scanNorm and scanNormExt classes to perform the normalisation.
        self._norm1 = scanNorm(
            scan=scan,
            norm_channel=norm_channel,
            norm_method=configChannel.normMethod.DIV,
            apply_to=apply_to,
        )
        self._norm2 = scanNormExt(
            scan=self._norm1,
            ext_scan=ext_scan,
            norm_channel=norm_channel,
            norm_method=configChannel.normMethod.MULT,
            apply_to=apply_to,
        )
        self._norm3 = scanNormExt(
            scan=self._norm2,
            ext_scan=ext_scan,
            norm_channel=double_norm_channel,
            norm_method=configChannel.normMethod.DIV,
            apply_to=apply_to,
        )
        # Copy the pointers to the final normalised data to this object.
        self._x = self._norm3.x
        self._x_errs = self._norm3.x_errs
        self._y = self._norm3.y
        self._y_errs = self._norm3.y_errs

        # Perform normalisation
        # self.load_and_normalise()
        return

    @property
    def ext_scan(self) -> scanAbstract:
        """
        The external scan object used for normalisation.

        Returns
        -------
        scanAbstract
            The external scan object.
        """
        return self._ext_scan

    @ext_scan.setter
    def ext_scan(self, scan: scanAbstract) -> None:
        self._ext_scan = scan
        # Propogate to normalisation objects
        self._norm2._ext_scan = scan
        self._norm3._ext_scan = scan

    @override
    def load_and_normalise(self) -> NoneType:
        # Need to sequentially load and normalise each step.
        self._norm1.load_and_normalise()
        self._norm2.load_and_normalise()
        self._norm3.load_and_normalise()
        # Copy the pointers to the final normalised data to this object.
        self._x = self._norm3.x
        self._x_errs = self._norm3.x_errs
        self._y = self._norm3.y
        self._y_errs = self._norm3.y_errs

    @override
    def _load_from_origin(self) -> NoneType:
        # Call the three normalisation steps to refresh data.
        raise NotImplementedError(
            "Double normalisation `_load_from_origin` method will not function as intended, \
                                  as 3 normalisation steps require sequential load and normalisation application."
        )

    @override
    def _apply_normalisation(self) -> None:
        raise NotImplementedError(
            "Double normalisation `_apply_normalisation` method will not function as intended, \
                                  as 3 normalisation steps require sequential load and normalisation application."
        )

    @override
    def _config_class(self) -> type[configBase]:
        """
        Return the settings class for the normalisation configuration.

        Required to allows loading from a settings object.

        Returns
        -------
        type[configChannel]
            The configuration class for the normalisation settings.
        """
        raise NotImplementedError(
            "Double normalisation configuration class not yet implemented."
        )
        return configBase  # TODO: Implement a double normalisation configuration class.

    @override
    @staticmethod
    def from_config(scan: scanAbstract, config: configBase) -> scanDoubleNorm:
        raise NotImplementedError(
            "Double normalisation from_config method not yet implemented."
        )

    @scanAbstractNorm.settings.getter
    @override
    def settings(self) -> configBase:
        raise NotImplementedError(
            "Double normalisation settings property not yet implemented."
        )

    @override
    def copy(self, *args, **kwargs) -> Self:
        clone = self.__class__(
            scan=self._origin,
            ext_scan=self._norm2._ext_scan,
            norm_channel=self._orig_norm_channel,
            double_norm_channel=self._orig_double_norm_channel,
            apply_to=self._apply_to,
            *args,
            **kwargs,
        )
        return clone


if __name__ == "__main__":
    import os
    import matplotlib.pyplot as plt
    import pyNexafs

    # Create a basic scan object form test data
    MEX2_MDA_PATH = os.path.abspath(
        os.path.normpath(
            os.path.join(
                # os.path.dirname(__file__),
                os.path.dirname(pyNexafs.__file__),
                "..\\tests\\test_data\\au\\MEX2\\2024-03\\MEX2_5640.mda",
            )
            .replace("\\", os.sep)
            .replace("/", os.sep)
        )
    )
    SXR_MDA_PATH = (
        os.path.abspath(
            os.path.normpath(
                os.path.join(
                    # Use package root
                    os.path.dirname(pyNexafs.__file__),
                    "..\\tests\\test_data\\au\\SXR\\2024-03\\sxr129598.mda",
                )
            )
        )
        .replace("\\", os.sep)
        .replace("/", os.sep)
    )
    from pyNexafs.parsers.au import MEX2_NEXAFS, SXR_NEXAFS

    # MEX2 example
    parser_mex2 = MEX2_NEXAFS(
        MEX2_MDA_PATH, relabel=True, energy_bin_domain=(3300, 3700)
    )
    scan_mex2 = parser_mex2.to_scan()

    # SXR Oxygen example
    parser_sxr = SXR_NEXAFS(SXR_MDA_PATH, relabel=True)
    scan_sxr = parser_sxr.to_scan()
    assert scan_sxr.y_labels[2] == "I0 VF", scan_sxr.y_labels[2]

    examples = [
        {
            "scan": scan_mex2,
            "norm_channel": "I0",
            "pre_edge_domain": (2460, 2465),
            "post_edge_domain": (2520, 2530),
            "idx": 1,
        },
        {
            "scan": scan_sxr,
            "norm_channel": scan_sxr.y_labels[2],
            "pre_edge_domain": (515, 525),
            "post_edge_domain": (570, 580),
            "idx": 0,
        },
    ]

    # plt.subplots(1,2, figsize=(10,5))
    # plt.plot(scan_sxr.x, scan_sxr.y[:,2], label="I0")
    # plt.show()

    for ex in examples[1:]:
        scan = ex["scan"]
        norm_channel = ex["norm_channel"]
        pre_edge_domain = ex["pre_edge_domain"]
        post_edge_domain = ex["post_edge_domain"]
        idx = ex["idx"]

        # Normalise to a channel
        norm = scanNorm(scan, norm_channel=norm_channel)

        # Normalise to the edges
        edge_c = configEdges(
            scan, pre_edge_domain=pre_edge_domain, post_edge_domain=post_edge_domain
        )
        edge_l = configEdges(
            scan,
            pre_edge_domain=pre_edge_domain,
            post_edge_domain=post_edge_domain,
            pre_edge_normalisation=edgeNormPre.LINEAR,
        )

        edge_e = configEdges(
            scan,
            pre_edge_domain=pre_edge_domain,
            post_edge_domain=post_edge_domain,
            pre_edge_normalisation=edgeNormPre.EXPONENTIAL,
        )

        # Plot each normalisation step
        fig, ax = plt.subplots(1, 3, sharex=True, figsize=(15, 5))
        ax[0].plot(
            scan.x,
            scan.y[:, idx],
            label=f"{type(scan_mex2).__name__}\n{scan.y_labels[idx]}",
        )

        line = ax[1].plot(
            norm.x, norm.y[:, idx], label=f"{type(norm).__name__}\n{norm.y_labels[idx]}"
        )
        ax[1].plot(
            scan.x,
            scan.y[:, idx],
            "--",
            label=f"{type(scan_mex2).__name__}\n{scan.y_labels[idx]}",
        )  # c=l[0].get_color()

        line = ax[2].plot(
            edge_c.x,
            edge_c.y[:, idx],
            label=f"{type(edge_c).__name__}\nConstant\n{edge_c.y_labels[idx]}",
        )
        line = ax[2].plot(
            edge_l.x,
            edge_l.y[:, idx],
            label=f"{type(edge_l).__name__}\nLinear\n{edge_l.y_labels[idx]}",
        )
        line = ax[2].plot(
            edge_e.x,
            edge_e.y[:, idx],
            label=f"{type(edge_e).__name__}\nExponential\n{edge_e.y_labels[idx]}",
        )
        ax[1].plot(
            edge_e.x,
            configEdges.EXP_FN_OFFSET(
                edge_e.x, *edge_e.pre_edge_fit_params[3 * idx : 3 * (idx + 1)]
            ),
            "--",
            c=line[0].get_color(),
            label="Exponential Subtraction",
        )

        ax[2].set_xlabel("Energy (eV)")
        for a in ax:
            a.set_ylabel("Intensity (a.u.)")
            a.legend()

        ax[0].set_title("Original")
        ax[1].set_title("Normalised to I0")
        ax[2].set_title("Normalised to Edges")
        # ax[2].set_yscale("log")
        fig.tight_layout()
        plt.show()



"""
A module for all normalisation and background subtraction operations on the data.

Normalisation operations are designed to be repeatable and tracable, such that
the order of operations is clear and the normalisation can be reproduced between
different datasets.
"""

# Config Classes
from pyNexafs.nexafs.normalisation.norm_settings import (
    configBase,
    configSeries,
    configChannel,
    configExternalChannel,
    configEdges,
)

# Config Enumerates
from pyNexafs.nexafs.normalisation.norm_settings import (
    normMethod,
    extSelection,
)

#
from pyNexafs.nexafs.normalisation.scan_normalised import (
    scanNorm,
    scanNormExt,
    scanDoubleNorm,
    scanNormEdges,
)

# from pyNexafs.nexafs.normalisation.scan_normalised import (
#     scan_norm,
#     scan_normalised_edges,
#     # scan_background_subtraction,
#     scan_normalised_background_channel,
# )

__all__ = [
    # Config Classes
    "configBase",
    "configSeries",
    "configChannel",
    "configExternalChannel",
    "configEdges",
    # Enumerates
    "normMethod",
    "extSelection",
    # Normalisation Classes
    "scanNorm",
    "scanNormExt",
    "scanDoubleNorm",
    "scanNormEdges",
]



import sys
from typing import override
import numpy as np
import matplotlib as mpl
import matplotlib.figure
import matplotlib.colors
from PyQt6 import QtGui, QtWidgets, QtCore
import PIL.ImageQt as PILImageQt
import PIL.Image as PILImage

mpl.use("QtAgg")


class QColormapPushButton(QtWidgets.QPushButton):
    def __init__(self, cmap, min_h=5, min_w=128, parent=None):
        super().__init__(parent=parent)
        # Set minimum size requirements
        self._min_h = min_h
        self._min_w = min_w
        button_min_height = self._min_h + 10
        button_min_width = self._min_w + 10
        self.setMinimumSize(button_min_height, button_min_width)
        # Set cmap
        self._pixmap = None
        self._cmap = None  # Initialise, before setter can raise an error.
        self.cmap = cmap  # Set cmap.
        # Set initial sizing
        self._w = self.width()
        self._h = self.height()

    @override
    def resizeEvent(self, event):
        if callable(self._resizeFn):
            self._resizeFn(event)

    def _resizeFn(self, event):
        if self.cmap is not None and event:
            # If there is a change in width, update the icon.
            w = self.width()
            if self._w != w:  # new width value.
                ### Check raw pixelwith of icon.
                if self._pixmap is not None and self._pixmap.width() < w:
                    # Generate new icon with double width.
                    self.generate_cmap_icon(w=2 * w, h=self._min_h)
                self.setIconSize(QtCore.QSize(self.width() - 20, self.height() - 10))

            # Update width / height values.
            self._w = w
            self._h = self.height()

    @property
    def cmap(self) -> matplotlib.colors.Colormap:
        return self._cmap

    @cmap.setter
    def cmap(self, val: matplotlib.colors.Colormap | str) -> None:
        if isinstance(val, str):
            if val not in matplotlib.colormaps:
                raise
            else:
                self._cmap = val
        elif isinstance(val, matplotlib.colors.Colormap):
            self._cmap = val
        else:
            raise TypeError(f"{val} is not a registered colormap string or object.")
        self.generate_cmap_icon(w=self._min_w, h=self._min_h)

    def generate_cmap_icon(self, w, h) -> None:
        qpix = self.cmap_to_qpix(cmap=self.cmap, pix_width=w, pix_height=h)
        self._pixmap = qpix
        icon = QtGui.QIcon(qpix)
        self.setIcon(icon)

    @staticmethod
    def cmap_to_qpix(
        cmap: matplotlib.colors.Colormap | str,
        pix_width: int = 128,
        pix_height: int = 20,
    ) -> QtGui.QPixmap:
        """
        Converts a matplotlib.colormap to a QtGui.QPixmap, for easy use in graphic elements.

        When the colormap has less elements than pix_width, the colormap is repeated to extend
        unique colours. When the colormap has more elements than pix_width (including from
        the result of repeating) uses interpolation to generate intermediate colours.

        Parameters
        ----------
        cmap : matplotlib.colors.Colormap
            A mapltolib colormap object, such as matplotlib.colormaps['magma']
        pix_height : int, optional
            The desired height of the QPixmap, by default 20 (pixels)
        pix_width : int, optional
            The desired width of the QPixmap, by default 128 (pixels)

        Returns
        -------
        QtGui.QPixmap
            A Qt pixelmap object, usable for graphic displays in GUIs.
        """
        # Get colours.
        if isinstance(cmap, str):
            cmap = mpl.colormaps[cmap]
        elif isinstance(cmap, matplotlib.colors.Colormap):
            pass
        else:
            raise TypeError(
                f"{cmap} is not a matplotlib Colormap or registered colormap name."
            )

        clrs = cmap(np.linspace(0, 1, pix_width))
        # Generate Qt Pixmap
        shape = (pix_height, *clrs.shape)
        data = np.zeros(shape)
        data[:] = clrs[np.newaxis] * 255
        pil_img = PILImage.fromarray(np.uint8(data), mode="RGBA")
        pil_imgqt = PILImageQt.ImageQt(pil_img)
        qpix = QtGui.QPixmap.fromImage(pil_imgqt)
        return qpix


class StyleSelector(QtWidgets.QWidget):
    def __init__(self, parent=None, cm_width=128, cm_height=20):
        super().__init__(parent)

        # Adjustable scroll
        scroll = QtWidgets.QScrollArea(self)
        scroll.setWidgetResizable(True)
        scroll.setHorizontalScrollBarPolicy(
            QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        scrollContent = QtWidgets.QWidget(scroll)
        scrollLayout = QtWidgets.QVBoxLayout()
        scrollLayout.addWidget(scroll)
        self.setLayout(scrollLayout)

        layout = QtWidgets.QGridLayout()
        scrollContent.setLayout(layout)
        scroll.setWidget(scrollContent)
        headers = ["Density", "Colourmap", "Bar"]
        for col, header in enumerate(headers):
            layout.addWidget(QtWidgets.QLabel(header + ":"), 0, col)
        layout.setColumnStretch(2, 1)

        # self.style_selector = QtWidgets.QComboBox()
        cmaps = list(matplotlib.colormaps)
        for n, c in enumerate(cmaps):
            cmap = mpl.colormaps[c]
            # Setup button
            button = QColormapPushButton(cmap, cm_width, cm_height)

            # Setup elements
            label_name = QtWidgets.QLabel()
            label_name.setText(c)
            label_density = QtWidgets.QLabel()
            label_density.setText(str(cmap.N))

            # Set scaling and expanding policy on elements:
            for elem in [button]:
                elem.setSizePolicy(
                    QtWidgets.QSizePolicy.Policy.Expanding,
                    QtWidgets.QSizePolicy.Policy.Preferred,
                )

            layout.addWidget(label_density, n + 1, 0)
            layout.addWidget(label_name, n + 1, 1)
            layout.addWidget(button, n + 1, 2)


# Demo of QT app
if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    window = StyleSelector()
    window.setWindowTitle("pyNexafs Style Selector")
    window.show()
    sys.exit(app.exec())



from PyQt6 import QtGui, QtWidgets, QtCore
import numpy as np
from pyNexafs.gui.widgets.graphing.matplotlib.graphs import FigureCanvas
import matplotlib.pyplot as plt
import matplotlib.colors as mpl_colors
import matplotlib as mpl
from matplotlib.backend_bases import MouseButton

# from pyNexafs.gui.widgets.graphing.matplotlib.widgets import NSpanSelector
from matplotlib.widgets import SpanSelector
from pyNexafs.gui.widgets.graphing.matplotlib.graphs import NavTBQT
import numpy.typing as npt
from pyNexafs.utils.mda import MDAFileReader
import os
from pyNexafs.utils.reduction import reducer


class EnergyBinReducer(QtWidgets.QWidget):
    """
    Reduces detector data with multiple energy bins (and channels) to a sum of counts.

    Parameters
    ----------
    energies : npt.NDArray
        A 1D array of energy values for the scanning beam energy.
    dataset : npt.NDArray
        The NEXAFS dataset with dimensions (beam_energy, detection_energy_bin, detector_index).
        The detector_index dimension is optional, used for multiple multi-channel analysers.
    bin_energies : npt.NDArray, optional
        A 1D array of energy values for each detection energy bin. If not provided, the bin numbers are used.
        If 2D, the second dimension must match the number of detectors.
    subsampling : int, optional
        The number of energy bins to skip when plotting the detector data.
        By default, every 5th bin is plotted.
    """

    def __init__(
        self,
        energies: npt.NDArray,
        dataset: npt.NDArray,
        bin_energies: npt.NDArray | None = None,
        subsampling: int = 5,
        parent=None,
    ):
        # Initialise the reduction class
        self._reduction: reducer = reducer(
            energies=energies, dataset=dataset, bin_energies=bin_energies
        )
        self._subsampling: int = subsampling

        # Initialize the parent class and layout
        super().__init__(parent)
        self._layout = QtWidgets.QVBoxLayout()
        self.setLayout(self._layout)
        if parent is not None:
            self.setContentsMargins(0, 0, 0, 0)
            self._layout.setContentsMargins(0, 0, 0, 0)

        # Generate a figure
        # For each detector, add a plot and colorbar
        root = int(np.ceil(np.sqrt(dataset.shape[2])))
        r, c = root, 2 * root
        subplots = plt.subplots(
            nrows=r,
            ncols=c,
            width_ratios=[1, 0.05] * r,
            layout="compressed",
        )
        self.detector_fig: plt.Figure = subplots[0]
        self.detector_axes: list[plt.Axes] = subplots[1]
        self.detector_canvas = FigureCanvas(self.detector_fig)
        self.detector_navtoolbar = NavTBQT(self.detector_canvas)
        self._layout.addWidget(self.detector_navtoolbar)
        self._layout.addWidget(self.detector_canvas)
        if len(self.detector_axes.shape) > 1:
            self.detector_axes = self.detector_axes.flatten()

        # Generate a second figure for the energy bin selection
        self._bin_layout = QtWidgets.QHBoxLayout()
        self._bin_edit_layout = QtWidgets.QVBoxLayout()
        self._bin_layout.addLayout(self._bin_edit_layout)
        self._bin_edit_layout.addWidget(QtWidgets.QLabel("Energy Binning:"))
        binning_label = " #:" if not self.reducer.has_bin_energies else " Energy (eV):"
        self._bin_edit_layout.addWidget(QtWidgets.QLabel("Lower" + binning_label))
        self.bin_lower = QtWidgets.QLineEdit()
        self.bin_lower.setMaximumWidth(100)
        self._bin_edit_layout.addWidget(self.bin_lower)
        self._bin_edit_layout.addWidget(QtWidgets.QLabel("Upper" + binning_label))
        self.bin_upper = QtWidgets.QLineEdit()
        self.bin_upper.setMaximumWidth(100)
        self._bin_edit_layout.addWidget(self.bin_upper)
        self.bin_lower.setValidator(
            QtGui.QDoubleValidator()
            if self.reducer.has_bin_energies
            else QtGui.QIntValidator()
        )
        self.bin_upper.setValidator(
            QtGui.QDoubleValidator()
            if self.reducer.has_bin_energies
            else QtGui.QIntValidator()
        )
        self._bin_edit_layout.addStretch(1)
        self._bin_edit_layout.addWidget(
            QtWidgets.QLabel(
                "Data shown has been modified\n to highlight features:\n"
                + "  1. Subtracted mean of each detector to\n    remove background noise levels.\n"
                + "  2. Translated from any negative values\n    to positive definite.\n"
                + "  3. Added 1e-2 to all values to allow \n    log plotting."
            )
        )

        self.bin_fig: plt.Figure
        self.bin_axes: plt.Axes
        self.bin_fig, self.bin_axes = plt.subplots(nrows=1, ncols=1)
        self.bin_canvas = FigureCanvas(self.bin_fig)
        self.bin_navtoolbar = NavTBQT(self.bin_canvas)
        self._bin_edit_layout.insertWidget(0, self.bin_navtoolbar)
        self._bin_layout.addWidget(self.bin_canvas)
        self._layout.addLayout(self._bin_layout)

        # Setup the energy bin selection
        self.plot_bin_features()

        # Setup the span selector
        self.span: SpanSelector = SpanSelector(
            # N=1,
            ax=self.bin_axes,
            onselect=self.onselect,
            direction="horizontal",
            useblit=True,
            props=dict(alpha=0.2, facecolor="red"),
            interactive=True,
            drag_from_anywhere=True,
            button=MouseButton.LEFT,
        )
        # plot the full range
        self.plot()

        # Connect the bounds to the span selector
        self.bin_lower.editingFinished.connect(self.on_bounds_update)
        self.bin_upper.editingFinished.connect(self.on_bounds_update)

    @property
    def reducer(self) -> reducer:
        """
        Returns the reducer object.

        Returns
        -------
        reducer
            The reducer object.
        """
        return self._reduction

    def plot_bin_features(self) -> None:
        """
        Plots the bin features for each detector.

        The bin features are the sum of counts for each detector channel.
        """
        self.bin_axes.clear()
        # Store the bins, so the max range can be used as a default.
        self._bins, counts = self.reducer.reduce_to_bin_features()

        for i in range(self.reducer.detectors):
            self.bin_axes.plot(self._bins, counts[:, i], label=f"Detector {i}")

        self.bin_axes.set_xlabel(
            "Bin #" if not self.reducer.has_bin_energies else "Bin Energy (eV)"
        )
        self.bin_axes.set_ylabel("Log Channel Counts (A.U.)")
        self.bin_axes.set_yscale("log")
        self.bin_axes.legend()
        self.bin_fig.tight_layout()

    def on_bounds_update(self):
        """
        Callback function for when the bounds are updated.

        Modifies span selection to match the new bounds,
        which will also trigger onselect.
        """
        lb, ub = self.bin_lower.text(), self.bin_upper.text()
        try:
            if self.reducer.has_bin_energies:
                lb, ub = float(lb), float(ub)
            else:
                lb, ub = int(lb), int(ub)
            self.span.extents = (lb, ub)
        except ValueError:
            # Do nothing if the values are invalid.
            pass

    @property
    def dataset(self) -> npt.NDArray:
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: npt.NDArray):
        self._dataset = dataset

        self.plot()

    @staticmethod
    def _validify_inputs(
        energies: npt.NDArray,
        dataset: npt.NDArray,
        bin_energies: npt.NDArray | None = None,
    ) -> bool:
        """
        Validates the input dataset, energies and bin_energies.

        Raises a ValueError if the inputs are invalid, otherwise returns True.

        Parameters
        ----------
        dataset : npt.NDArray
            The NEXAFS dataset with dimensions (beam_energy, detection_energy_bin, detector_index).
            The detector_index dimension is optional, used for multiple multi-channel analysers.
        energies : npt.NDArray
            A 1D array of energy values for the scanning beam energy.
        bin_energies : npt.NDArray, optional
            A 1D array of energy values for each detection energy bin. If not provided, the bin numbers are used.
            If 2D, the second dimension must match the number of detectors.

        Returns
        -------
        tuple[npt.NDArray, npt.NDArray, npt.NDArray]
            The validated dataset, energies and bin_energies.
        """
        if not energies.ndim == 1:
            raise ValueError("`energies` must be a 1D numpy array.")
        if not (dataset.ndim == 2 or dataset.ndim == 3):
            raise ValueError(
                "`dataset` must be a 3D or 2D numpy array with indexes corresponding to"
                + " (beam_energy, detection_energy_bin, detector_index). The detector_index dimension is optional."
            )
        if not (energies.shape[0] == dataset.shape[0]):
            raise ValueError(
                f"`energies` must have the same length ({energies.shape[0]})"
                + f" as the first dimension of `dataset` ({dataset.shape[0]})."
            )
        if dataset.ndim == 2:
            dataset = dataset[:, :, np.newaxis]  # add final dimension.

        # Check bin_energies validity
        if bin_energies is not None:
            bin_energies = np.asarray(bin_energies)
            if not (bin_energies.ndim == 1 or bin_energies.ndim == 2):
                raise ValueError(
                    "`bin_energies` must be a 1D or 2D numpy array, with indexes corresponding to"
                    + " (detection_bin_energy, detector_index). The second dimension is optional."
                )
            if not (bin_energies.shape[0] == dataset.shape[1]):
                raise ValueError(
                    f"`bin_energies` must have the same length ({bin_energies.shape[0]}) "
                    + f"as the second dimension of `dataset` ({dataset.shape[1]})."
                )
        return True

    @property
    def domain(self) -> tuple[float, float] | tuple[int, int]:
        """
        The selection domain.

        Energies (`float`) are returned if `bin_energies` have been provided,
        otherwise bin indexes (`int`) are used.

        Returns
        -------
        lower : float | int
            The lower bound of the selected domain.
        upper : float | int
            The upper bound of the selected domain.
        """
        vals = (
            self.span.extents
            if self.reducer.has_bin_energies
            else tuple(np.round(self.span.extents).astype(int))
        )
        if vals[0] == vals[1]:
            # If the same value is provided, return the full range from the data.
            return tuple(self.reducer.bin_energies[[0, -1]])
        else:
            return vals

    @property
    def domain_incidies(self) -> list[tuple[int, int]]:
        """
        A list of the bin indices for each detector.

        Returns
        -------
        list[tuple[int, int]]
            List of the lower and upper selected bin indices for each detector.
        """
        return self.reducer.domain_to_index_range(self.domain)

    def plot(self, bin_domain: tuple[int, int] | None = None):
        """
        Plot the detector data for the given energy domain.

        Parameters
        ----------
        bin_domain : tuple[int, int], optional
            The energy domain to plot, by default None.
            If None, plots the full range.
        """
        # Setup the colormap to plot the energies
        cmap = plt.get_cmap("inferno")

        # Clear all existing axes (plots and colorbars)
        for ax in self.detector_axes:
            ax.clear()

        # Get the sub-sampled data from the dataset
        sampled_bin_energies, sampled_data = self.reducer.reduce_domain(
            bin_domain=bin_domain
        )

        # Plot the new data at the selected domain
        for i, ax in enumerate(self.detector_axes[::2]):  # for each detector
            # Collect the detector data
            energies = self.reducer.energies
            bin_energies = (
                sampled_bin_energies[:: self._subsampling, i]
                if sampled_bin_energies.ndim == 2
                else sampled_bin_energies
            )
            data = (
                sampled_data[:, :: self._subsampling, i]
                if sampled_data.ndim == 3
                else sampled_data
            )

            # Setup the color normalisation.
            norm = mpl_colors.Normalize(
                vmin=bin_energies.min(), vmax=bin_energies.max()
            )

            # Plot the data
            for j in range(data.shape[1]):  # for each subsampled bin
                ax.plot(
                    energies,
                    data[:, j],
                    c=cmap(norm(bin_energies[j])),
                )

            # Add the colorbar
            mpl.colorbar.ColorbarBase(
                self.detector_axes[2 * i + 1],
                cmap=cmap,
                norm=norm,
                label=("Bin Energy (eV)" if self.reducer.has_bin_energies else "Bin #"),
            )

            ax.set_xlabel("Energy (eV)")
            ax.set_ylabel("Counts")
            ax.set_title("Detector {}".format(i))
        # Update the canvas
        self.detector_canvas.draw()

    def onselect(self, xmin: float, xmax: float):
        """
        Callback function for the SpanSelector.

        Used to call `plot` with the selected energy domain.
        If upper and lower bounds are not provided / identical,
        then the full range is used.

        Parameters
        ----------
        xmin : float
            The lower bound of the selected energy domain.
        xmax : float
            The upper bound of the selected energy domain.
        """
        # If the same value is provided, or range is below visibility.
        if xmin == xmax or abs(xmax - xmin) < self.span.minspan:
            # Return the full range from the data.
            self.plot(bin_domain=None)
        else:
            self.plot(bin_domain=(xmin, xmax))

        # Pause QEdit updates while changing values from the span selector
        self.bin_lower.blockSignals(True)
        self.bin_upper.blockSignals(True)
        if self.reducer.has_bin_energies:
            self.bin_lower.setText(f"{xmin:.6f}")
            self.bin_upper.setText(f"{xmax:.6f}")
        else:
            self.bin_lower.setText(str(int(round(xmin))))
            self.bin_upper.setText(str(int(round(xmax))))
        # Resume QEdit updates
        self.bin_lower.blockSignals(False)
        self.bin_upper.blockSignals(False)


class EnergyBinReducerDialog(QtWidgets.QDialog):
    def __init__(
        self,
        energies: npt.NDArray,
        dataset: npt.NDArray,
        bin_energies: npt.NDArray | None = None,
        subsampling: int = 5,
        parent=None,
    ):
        super().__init__(parent=parent)

        self.reducerUI = EnergyBinReducer(
            energies, dataset, bin_energies, subsampling, parent=self
        )
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.addWidget(self.reducerUI)
        self.setLayout(self._layout)
        self.setWindowTitle("Energy Bin Reducer")

        # self.dataset = dataset
        # self.layout = QtWidgets.QVBoxLayout()
        # self.setLayout(self.layout)
        # self.energy_bin_reducer = EnergyBinReducer(dataset=dataset)
        # self.layout.addWidget(self.energy_bin_reducer)

        # Add buttons
        QBtn = (
            QtWidgets.QDialogButtonBox.StandardButton.Cancel
            | QtWidgets.QDialogButtonBox.StandardButton.Ok
        )
        self._buttonBox = QtWidgets.QDialogButtonBox(QBtn)
        self._buttonBox.accepted.connect(self.accept)
        self._buttonBox.rejected.connect(self.reject)
        self._layout.addWidget(self._buttonBox)

    @property
    def reducer(self) -> reducer:
        """
        Returns the reducer object.

        Returns
        -------
        reducer
            The reducer object.
        """
        return self.reducerUI.reducer

    @property
    def domain_indices(self) -> list[tuple[int, int]]:
        """
        Returns the selected energy domain indices for each detector.

        Returns
        -------
        list[tuple[int, int]]
            List of the lower and upper selected bin indices for each detector.
        """
        return self.reducerUI.domain_incidies

    @property
    def domain(self) -> tuple[float, float] | tuple[int, int]:
        """
        Returns the selected energy domain.

        Returns
        -------
        tuple[float, float] | tuple[int, int]
            The lower and upper selected energies/indexes.
            If energies, a tuple of floats is returned, otherwise a tuple of integers for indices.
        """
        return self.reducerUI.domain

    @property
    def result_sum(self) -> tuple[npt.NDArray, npt.NDArray]:
        """
        Returns the sum of counts for the selected energy domain.

        Returns
        -------
        tuple[npt.NDArray, npt.NDArray]
            The sum of counts for the selected energy domain.
            The first array is the sum of counts for each detector.
            The second array is the sum of counts for all detectors.
        """
        return self.reducer.reduce_by_sum(bin_domain=self.domain)

    def accept(self):
        # Check if all settings are valid
        # valid, message = self.validate_settings()
        valid = True  # No validation required for simple selector..
        if not valid:
            dlg = QtWidgets.QMessageBox(self)
            dlg.setWindowTitle("Normalisation Error")
            dlg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
            dlg.setIcon(QtWidgets.QMessageBox.Icon.Warning)
            dlg.setText("Error in normalisation settings.")
            dlg.exec()
        else:
            super().accept()

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        if event.key() == QtCore.Qt.Key.Key_Escape:
            self.reject()
        elif (
            event.key() == QtCore.Qt.Key.Key_Return
            or event.key() == QtCore.Qt.Key.Key_Enter
        ):
            if (
                self.reducerUI.bin_lower.hasFocus()
                or self.reducerUI.bin_upper.hasFocus()
            ):
                # Ignore the event when editing.
                return
            else:
                return event.accept()
        else:
            super().keyPressEvent(event)

    def reject(self):
        super().reject()


if __name__ == "__main__":
    path = os.path.dirname(__file__)
    package_path = os.path.normpath(os.path.join(path, "../../../"))
    mda_path = os.path.normpath(
        os.path.join(package_path, "tests/test_data/au/MEX2/MEX2_5643.mda")
    )
    mda_data, mda_scans = MDAFileReader(mda_path).read_scans()
    data1D, data2D = mda_data

    # Convert MEX2 Data
    energies = data1D[:, 0] * 1000  # Convert keV to eV
    span = [80, 900]
    binned_data = data2D[:, span[0] : span[1], :]
    TOTAL_BINS = span[1] - span[0]
    BIN_ENERGY = 11.935
    BIN_96_ENERGY = 1146.7
    BIN_I_ENERGY = BIN_96_ENERGY - (95 - span[0]) * BIN_ENERGY
    bin_energies = np.linspace(
        BIN_I_ENERGY, BIN_I_ENERGY + TOTAL_BINS * BIN_ENERGY, TOTAL_BINS
    )

    app = QtWidgets.QApplication([])
    win = EnergyBinReducerDialog(
        energies=energies, dataset=binned_data, bin_energies=bin_energies
    )
    win.show()
    val = win.exec()
    print(val)
    if val:
        print("OK")
        print("Selected Domain:", win.domain)
        print("Selected Domain Indices for detectors:", win.domain_indices)
    else:
        print("Cancel")



"""
Parser classes for the Medium Energy X-ray 2 (MEX2) beamline at the Australian Synchrotron.
"""

# Internal
from pyNexafs.parsers import parserBase
from pyNexafs.utils.mda import MDAFileReader
from pyNexafs.utils.reduction import reducer
from pyNexafs.parsers.au.aus_sync.MEX2_relabels import RELABELS
from pyNexafs.types import dtype, parse_fn_ret_type, reduction_type
from pyNexafs.parsers.au.aus_sync.MEX_detectors import (
    DanteFluorescence,
    Xpress3Fluorescence,
)
from pyNexafs.nexafs import scanBase

# Standard
import typing
import ast
import warnings
import datetime
import os
import io

# External
import numpy as np
import numpy.typing as npt

has_PYQT: bool
try:
    from PyQt6 import QtWidgets

    has_PYQT = True
    from pyNexafs.gui.widgets.reducer import EnergyBinReducerDialog
except ImportError:
    has_PYQT = False


def MEX2_datetime_conversion(MEX2_TIME: float) -> datetime.datetime:
    """
    Convert the MEX2 datetime format to a UTC datetime object.

    Note that for Australia, this will be +11 during daylight savings time,
    and +10 otherwise. This function does not account for this, so the
    returned datetime object will be in UTC time.

    Parameters
    ----------
    MEX2_TIME : float
        The MEX2 time in seconds since the epoch (2023-01-01).

    Returns
    -------
    datetime.datetime
        The converted datetime object in UTC.
    """
    # dtime19900101 = datetime.datetime(1990, 1, 1, 0, 0, 0, 0)  # MEX2 epoch
    dtime20230101 = datetime.datetime(2023, 1, 1, 0, 0, 0, 0)  # MEX2 epoch
    # Convert MEX2 time to seconds since 1990-01-01
    mex2_time = datetime.timedelta(seconds=MEX2_TIME)
    # Convert to datetime object
    ret_time = dtime20230101 + mex2_time
    return ret_time


class MEX2_NEXAFS(parserBase):
    """
    Australian Synchrotron Medium X-ray (MEX) NEXAFS parser.

    Parses data formats including '.xdr' and '.mda' formats from the SXR
    Near Edge X-ray Absorption Fine Structure (NEXAFS) tool.

    Parameters
    ----------
    filepath : str | None
        The file path to the data file.
    header_only : bool, optional
        If True, only the header of the file is read, by default False.
    relabel : bool | None, optional
        If True, then the parser will relabel the data columns, by default None.
    **kwargs
        Additional keyword arguments that will be passed to the `file_parser` method.

    Attributes
    ----------
    ALLOWED_EXTENSIONS
    SUMMARY_PARAM_RAW_NAMES
    COLUMN_ASSIGNMENTS
    RELABELS

    Notes
    -----
    Implemented for data as of 2024-Mar.
    """

    ALLOWED_EXTENSIONS: list[str] = [".xdi", ".mda"]
    SUMMARY_PARAM_RAW_NAMES: list[str | tuple[str, ...]] = [
        ("Sample", "Comment 1", "MEX2SSCAN01:saveData_comment1"),
        "Angle",
        "Element.symbol",
        "Element.edge",
        "E1",
        "E2",
        "E3",
        "E4",
    ]
    COLUMN_ASSIGNMENTS = {
        "x": ("Energy", "Energy Setpoint", "energy"),
        "y": [
            ("I0", "i0"),
            ("Sample Drain", "SampleDrain"),
            (
                "Fluorescence",
                "ifluor",
                "Fluorescence Sum",
                "Fluorescence Sum (Reduced)",
            ),
            ("Gate Time Setpoint", "Count Time", "count_time"),
            "Fluorescence Detector 1",
            "Fluorescence Detector 2",
            "Fluorescence Detector 3",
            "Fluorescence Detector 4",
            ("Bragg", "bragg"),
        ],
        "y_errs": None,
        "x_errs": None,
    }

    # See MEX2_relabels.py for the relabels dictionary.
    RELABELS = RELABELS

    # Mappings
    CHANNEL_MAP = {
        "MEX2ES01ZEB01:CALC_ENERGY_EV": dtype.E,  # measured energy
        "MEX2ES01DAQ01:ch2:S:MeanValue_RBV": dtype.I0,
        "MEX2ES01DAQ01:ch3:S:MeanValue_RBV": dtype.TEY,
        "ifluor": dtype.PFY,
    }

    def __init__(
        self,
        filepath: str | None,
        header_only: bool = False,
        relabel: bool | None = None,
        **kwargs,
    ) -> None:
        # Could do something with other arguments here...
        super().__init__(filepath, header_only, relabel, **kwargs)

    @typing.override
    def reduce(
        self,
        use_prior_params: bool = True,
        energy_bin_domain: tuple[float, float] | None = None,
    ) -> reduction_type:
        """
        Perform data reduction on the loaded data, particularly relevant for .mda files with Fluorescence binning.

        Parameters
        ----------
        use_prior_params : bool, optional
            Previously used parameters are stored on the class, matching the parser fn.
        energy_bin_domain : tuple[float, float] | None, optional
            The energy domain (in eV) to bin the data, by default None.

        Returns
        -------
        reduced_data : npt.NDArray | None
            The reduced data array, or None if no reduction is performed.
        reduced_labels : list[str | None] | None
            The reduced labels, or None if no reduction is performed.
        reduced_units : list[str | None] | None
            The reduced units, or None if no reduction is performed.
        """
        parser_fn = self._parser_fn
        if parser_fn is None:
            return None, None, None

        match parser_fn:
            case self.parse_mda_2024_11 | self.parse_mda_2025_02:
                # Get the appropriate MCA channel names:
                if parser_fn == self.parse_mda_2024_11:
                    detector = DanteFluorescence
                elif parser_fn == self.parse_mda_2025_02:
                    detector = Xpress3Fluorescence
                else:
                    raise ValueError("Invalid parser function for MDA reduction.")

                # Reduce the data
                data = self.data
                labels = self.labels
                if data is None:
                    return None, None, None
                else:
                    if isinstance(data, tuple):
                        # Reduce the Flourescence data!
                        flour_data = data[1]
                        assert flour_data.ndim == 3, (
                            f"Fluorescence data should be indexed with energy, binning and channel but has rank {data.ndim}."
                        )
                        assert isinstance(labels, tuple) and len(labels) > 1, (
                            f"Fluorescence data labels not found, instead labels={labels}."
                        )
                        flour_labels = labels[1]

                        # Collect the labels
                        fluorescence_labels = detector.FLUOR_NAMES
                        relabelled_flour_labels = [
                            self.RELABELS[label] for label in fluorescence_labels
                        ]
                        # Check strings in labels match expected values, either in original or relabelled form.
                        if (
                            fluorescence_labels != flour_labels
                            and flour_labels != relabelled_flour_labels
                        ):
                            raise ValueError(
                                f"Fluorescence detector labels do not match expected values.\nFound {flour_labels}, expected {fluorescence_labels} or {relabelled_flour_labels}. Reduction failed."
                            )

                        # Take properties from 1D and 2D arrays:
                        energies = (
                            data[0][:, 0] * 1000
                        )  # Convert keV to eV for MEX beamline
                        interest_bins = detector.INTERESTING_BIN_IDX
                        bin_slice = (
                            slice(interest_bins[0], interest_bins[1])
                            if interest_bins is not None
                            else slice(None)
                        )
                        dataset = flour_data[:, bin_slice, :]  # eneergy, bins, channels
                        bin_e = detector.INTERESTING_BIN_ENERGIES()  # pre-calibrated.

                        ## Perform binning on 2D array:
                        # Is an existing binning range available?
                        reduction_kwargs = self.reduction_kwargs
                        if energy_bin_domain is not None:
                            # Update the class variable with the new binning settings.
                            red = reducer(energies, dataset, bin_e)
                        elif (
                            use_prior_params
                            and reduction_kwargs is not None
                            and "energy_bin_domain" in reduction_kwargs
                        ):
                            # Uses the most recent binning settings.
                            red = reducer(energies, dataset, bin_e)
                            energy_bin_domain = reduction_kwargs["energy_bin_domain"]
                        else:
                            if has_PYQT:
                                # Create a QT application to run the dialog.
                                if QtWidgets.QApplication.instance() is None:
                                    _ = QtWidgets.QApplication([])

                                # Run the Bin Selector dialog
                                window = EnergyBinReducerDialog(
                                    energies=energies,
                                    dataset=dataset,
                                    bin_energies=bin_e,
                                )
                                window.show()
                                if window.exec():
                                    # If successful, store the binning settings and data reducer
                                    energy_bin_domain = window.domain
                                    red = window.reducer
                                else:
                                    raise ValueError("No binning settings selected.")
                            else:
                                raise ValueError(
                                    "No binning settings selected, and no PyQt available to run the dialog."
                                )
                                # TODO: Add a command line interface for binning settings.

                        # Use the binning settings to reduce the data.
                        _, reduced_summed_data = red.reduce_by_sum(
                            bin_domain=energy_bin_domain,
                            axis=None,  # all axes
                        )

                        _, reduced_single_detector_data = red.reduce_by_sum(
                            bin_domain=energy_bin_domain,
                            axis="bin_energies",  # just the bin energies
                        )

                        # Concatenate the reduced data:
                        reduced_data = np.concatenate(
                            [
                                reduced_single_detector_data,
                                reduced_summed_data[:, np.newaxis],
                            ],
                            axis=-1,
                        )

                        # Save the energy bin domain
                        self.reduction_kwargs = dict(
                            energy_bin_domain=energy_bin_domain
                        )

                        # Return the reduced data.
                        return (
                            reduced_data,
                            detector.FLUOR_NAMES + ["ifluor"],
                            ["a.u."] * (reduced_data.shape[1]),
                        )

                    else:
                        warnings.warn("Data is not a tuple, no reduction performed.")
                        return None, None, None
            case _:
                # No reduction performed
                return None, None, None
        return super().reduce()

    @typing.override
    def to_scan(
        self,
        use_prior_params: bool = True,
        energy_bin_domain: tuple[float, float] | None = None,
        *,
        load_all_columns: bool = False,
        warn_missing_labels: bool = True,
        only_labels: bool = False,
        scan_obj: scanBase | None = None,
        **kwargs,
    ) -> scanBase:
        """
        Same as `parserBase.to_scan`, but with additional reduction parameters.

        Parameters
        ----------
        use_prior_params : bool, optional
            Previously used parameters are stored on the class, matching the parser fn.
        energy_bin_domain : tuple[float, float] | None, optional
            The energy domain (in eV) to bin the data, by default None.
        load_all_columns : bool, optional
            If True, then all columns in the data are loaded, by default False.
        warn_missing_labels : bool, optional
            If True, then a warning is raised if any expected labels are missing, by default True.
        only_labels : bool, optional
            If True, then only the labels are returned, by default False.
        scan_obj : scanBase | None, optional
            An existing scan object to populate, by default None.
        **kwargs
            Additional keyword arguments that will be passed to the `file_parser` method.

        Returns
        -------
        scanBase
            The populated scan object.
        """
        return super().to_scan(
            use_prior_params=use_prior_params,
            energy_bin_domain=energy_bin_domain,
            load_all_columns=load_all_columns,
            warn_missing_labels=warn_missing_labels,
            only_labels=only_labels,
            scan_obj=scan_obj,
            **kwargs,
        )

    @classmethod
    def parse_xdi(
        cls,
        file: typing.IO | str,
        header_only: bool = False,
    ) -> parse_fn_ret_type:
        """
        Read Australian Synchrotron '.xdi' files.

        Parameters
        ----------
        file : typing.IO | str
            An of the datafile (i.e. open('file.xdi', 'r')).
        header_only : bool, optional
            If True, then only the header of the file is read and npt.NDArray is returned as None, by default False.

        Returns
        -------
        tuple[NDArray, list[str], dict[str, Any]]
            Returns a set of data as a numpy array,
            labels as a list of strings,
            units as a list of strings,
            and parameters as a dictionary.

        Raises
        ------
        ValueError
            If the file is not a valid .xdi file.
        """
        # Ensure text and read mode:
        if isinstance(file, str):
            # Open the file in text mode
            with open(file, "r", encoding="utf-8") as f:
                return cls.parse_xdi(f, header_only=header_only)
        elif isinstance(file, typing.BinaryIO):
            return cls.parse_xdi(
                io.TextIOWrapper(file, encoding="utf-8"), header_only=header_only
            )
        else:
            assert isinstance(file, typing.TextIO), (
                "File must be a typing.IO implementation or file path string."
            )
        # Check read mode
        if "r" not in file.mode:
            raise ValueError(f"File {file.name} is not opened in read mode.")

        # Initialise structures
        params = {}
        labels = []
        units = None

        # Check valid format.
        if not file.name.endswith(".xdi"):
            raise ValueError(f"File {file.name} is not a valid .xdi file.")

        ### Read file
        # Check header is correct
        assert file.readline() == "# XDI/1.0\n", "Invalid XDI file header."

        ## 1 Initial Parameters
        column_descriptions = {}
        column_type_assignments = {}
        # Read first param line.
        line = file.readline()
        i = 0
        while "# " == line[0:2] and line != "# ///\n":
            line = line[2:].strip().split(": ", 1)  # split first colon
            param, value = tuple(line)

            # Categorise information
            if param.startswith("Column."):
                # Label name
                labels.append(value)
                params[param] = value
            elif param.startswith("Describe.Column."):
                # Label description. No dedicated structure for this, add to params.
                col_name = params[param.replace("Describe.", "")]
                if col_name + " - " in value:
                    value = value.replace(col_name + " - ", "")
                if " -- " in value:
                    value, descr = value.split(" -- ", 1)
                else:
                    descr = None
                params[param] = value
                column_descriptions[i] = (
                    value,
                    col_name + " -- " + descr if descr is not None else col_name,
                    "",  # No units
                )
                column_type_assignments[i] = f"1-D Detector{i:4}"
            else:
                if param == "Samples":
                    # Parse the datapoint length list
                    samples = ast.literal_eval(value)
                    samples = [x.strip() if isinstance(x, str) else x for x in samples]
                    # Add to params
                    for i in range(len(samples)):
                        params[f"Datapoints.Column.{i}"] = samples[i]
                else:
                    # General parameter, add to params.
                    params[param] = value
            # Load new param line
            line = file.readline()
            i += 1

        assert line == "# ///\n", (
            "End of initial parameter values"
        )  # Check end of initial parameters

        # Get samplename from first comment line
        line = file.readline()
        sample_name = line[2:].strip()
        params["Sample"] = sample_name
        params["Comment 1"] = sample_name

        # Read the second comment line:
        line = file.readline()
        if line != "# \n":
            comment2 = line[2:].strip()
            params["Comment 2"] = comment2

        # Some xdi conversion have a "default mda2xdi" line.
        line = file.readline()
        try:
            assert line == "# xdi from default mda2xdi preset for mex2.\n", (
                "Conversion line"
            )
            assert file.readline() == "#--------\n", "Conversion line"
        except AssertionError:
            assert line == "#--------\n", "Conversion line"

        # Read data columns
        header_line = file.readline()
        assert header_line[0:2] == "# ", "Start of data columns"
        labels = (
            header_line[2:].strip().split()
        )  # split on whitespace, even though formatting seems to use "   " (i.e. three spaces).
        labels = [label.strip() if type(label) is str else label for label in labels]

        if header_only:
            # Do not process remaining lines
            return None, labels, units, params

        # Read data
        lines = file.readlines()  # read remaining lines efficiently

        # Convert data to numpy array.
        data = np.loadtxt(lines)
        data = np.array(data)

        return data, labels, units, params

    @classmethod
    def parse_mda_2025_02(
        cls, file: typing.IO | str, header_only: bool = False
    ) -> parse_fn_ret_type:
        """
        Read Australian Synchrotron MEX2 NEXAFS '.mda' files.

        Created for data as of 2025-March. This is after the Fluorescence MCA detectors were updated
        from Dante to instead use Xpress3. The energy binning has also been changed to be in eV instead of keV.
        Note, when reducing this data, the fluorescence detector data is also corrected according to
        the input and output count rates (i.e. the deadtime of the detector).

        Parameters
        ----------
        file : typing.IO | str
            Implementation of typing.IO of the datafile (i.e. open('file.mda', 'rb')).
        header_only : bool, optional
            If True, then only the header of the file is read and
            npt.NDArray is returned as None, by default False.

        Returns
        -------
        data : npt.NDArray | tuple[npt.NDArray, ...] | None
            A set of data as a numpy array, or a tuple of numpy arrays for multiple scans.
        labels : list[str | None] | list[str] | tuple[list[str] | None, ...]
            Labels as a list of strings, or a tuple of lists for multiple scans.
        units : list[str | None] | list[str] | tuple[list[str] | None, ...]
            Units as a list of strings, or a tuple of lists for multiple scans.
        params : dict[str, Any]
            Parameters as a dictionary.

        Raises
        ------
        ValueError
            If the file is not a valid .mda file.
        """
        if isinstance(file, str):
            if not os.path.exists(file):
                raise ValueError(f"File path '{file}' cannot be found for MDA reading.")
            # Open the file in binary mode
            with open(file, "rb") as f:
                return cls.parse_mda_2025_02(f, header_only=header_only)
        elif isinstance(file, (typing.TextIO, io.TextIOWrapper)):
            return cls.parse_mda_2025_02(file.buffer, header_only=header_only)
        else:
            assert isinstance(file, (typing.BinaryIO, io.BufferedReader, io.BytesIO)), (
                f"File must be a typing.BinaryIO implementation or file path string. Was '{type(file)}'."
            )

        # Initialise parameter list
        params = {}
        data: npt.NDArray | tuple[npt.NDArray, ...] | None = None
        labels: list[str | None] | list[str] | tuple[list[str] | None, ...]
        units: list[str | None] | list[str] | tuple[list[str] | None, ...]

        # Check valid format.
        if not file.name.endswith(".mda"):
            raise ValueError(f"File {file.name} is not a valid .mda file.")

        # Need to reopen the file in byte mode.
        mda = MDAFileReader(file)
        mda_header = mda.read_header_as_dict()
        ## Previously threw error for higher dimensional data, now just a warning.
        mda_params = mda.read_parameters()
        mda_arrays, mda_scans = mda.read_scans(header_only=header_only)

        # Convert to tuple if multiple dimensional scans
        labels = tuple([[] for _ in range(len(mda_scans))])
        units = tuple([[] for _ in range(len(mda_scans))])

        # Add values to params dict
        params.update(mda_header)
        params.update(mda_params)

        positioners = None
        detectors = None
        for i, scan in enumerate(mda_scans):
            # Add new positioners and detectors to the existing lists
            positioners = (
                scan.positioners
                if positioners is None
                else positioners + scan.positioners
            )
            detectors = (
                scan.detectors if detectors is None else detectors + scan.detectors
            )

            # Check 'multi-channel-analyser-spectra of fluorescence-detector' names are as expected
            if i == 1:
                assert scan.labels() == Xpress3Fluorescence.FLUOR_NAMES, (
                    f"Fluorescence detector labels do not match expected values. Found {scan.labels()}"
                )

            scan_labels = labels[i]
            scan_units = units[i]
            for p in scan.positioners:
                scan_labels.append(p.name)
                scan_units.append(p.unit)
            for d in scan.detectors:
                scan_labels.append(d.name)
                scan_units.append(d.unit)

        if len(mda_scans) == 1:
            labels = labels[0]
            units = units[0]

        if mda_arrays is not None:
            data = tuple(mda_arrays)

        # Collect a full set of the column descriptions
        column_types = {
            "Positioner": (
                "name",
                "descr",
                "step mode",
                "unit",
                "rdbk name",
                "rdbk descr",
                "rdbk unit",
            ),
            "Detector": (
                "name",
                "descr",
                "unit",
            ),
        }

        column_descriptions = (
            {
                i: [
                    p.name,
                    p.desc,
                    p.step_mode,
                    p.unit,
                    p.readback_name,
                    p.readback_desc,
                    p.readback_unit,
                ]
                for i, p in enumerate(positioners)
            }
            if positioners is not None
            else {}
        )
        if detectors is not None:
            column_descriptions.update(
                {
                    i + (len(positioners) if positioners is not None else 0): [
                        d.name,
                        d.desc,
                        d.unit,
                    ]
                    for i, d in enumerate(detectors)
                }
            )
        params["column_types"] = column_types
        params["column_descriptions"] = column_descriptions

        # Use scan time if available, otherwise let system time be used.
        if "MEX1ES01GLU01:MEX_TIME" in params:
            assert (
                params["MEX1ES01GLU01:MEX_TIME"][0] == "Seconds since 1990.01.01"
            )  # HAHA this is actually 2023.01.01 :')
            newtime = MEX2_datetime_conversion(params["MEX1ES01GLU01:MEX_TIME"][2])
            params["created"] = newtime

        return data, labels, units, params

    @classmethod
    def parse_mda_2024_11(
        cls,
        file: typing.IO | str,
        header_only: bool = False,
    ) -> parse_fn_ret_type:
        """
        Read Australian Synchrotron MEX2 NEXAFS .mda files.

        Created for data as of 2024-Nov.

        Parameters
        ----------
        file : typing.IO | str
            Implementation of typing.IO or file path string of the datafile (i.e. open('file.mda', 'rb')).
        header_only : bool, optional
            If True, then only the header of the file is read and
            npt.NDArray is returned as None, by default False.

        Returns
        -------
        data : npt.NDArray | tuple[npt.NDArray, ...] | None
            A set of data as a numpy array, or a tuple of numpy arrays for multiple scans.
        labels : list[str | None] | list[str] | tuple[list[str] | None, ...]
            Labels as a list of strings, or a tuple of lists for multiple scans.
        units : list[str | None] | list[str] | tuple[list[str] | None, ...]
            Units as a list of strings, or a tuple of lists for multiple scans.
        params : dict[str, Any]
            Parameters as a dictionary.

        Raises
        ------
        ValueError
            If the file is not a valid .mda file.
        """
        if isinstance(file, str):
            if not os.path.exists(file):
                raise ValueError(f"File path '{file}' cannot be found for MDA reading.")
            # Open the file in binary mode
            with open(file, "rb") as f:
                return cls.parse_mda_2024_11(f, header_only=header_only)
        elif isinstance(file, (typing.TextIO, io.TextIOWrapper)):
            # use the underlying binary buffer
            return cls.parse_mda_2024_11(file.buffer, header_only=header_only)
        else:
            assert isinstance(file, (typing.BinaryIO, io.BufferedReader, io.BytesIO)), (
                "File must be a typing.BinaryIO implementation or file path string."
            )

        # Initialise parameter list
        params = {}
        data: npt.NDArray | tuple[npt.NDArray, ...] | None = None
        labels: list[str | None] | list[str] | tuple[list[str] | None, ...]
        units: list[str | None] | list[str] | tuple[list[str] | None, ...]

        # Check valid format.
        fname = file.name
        if not fname.endswith(".mda"):
            raise ValueError(f"File {fname} is not a valid .mda file.")

        mda = MDAFileReader(file)
        mda_header = mda.read_header_as_dict()
        ## Previously threw error for higher dimensional data, now just a warning.
        mda_params = mda.read_parameters()
        mda_arrays, mda_scans = mda.read_scans(header_only=header_only)

        # Convert to tuple if multiple dimensional scans
        labels = tuple([[] for _ in range(len(mda_scans))])
        units = tuple([[] for _ in range(len(mda_scans))])

        # Add values to params dict
        params.update(mda_header)
        params.update(mda_params)

        positioners = None
        detectors = None
        for i, scan in enumerate(mda_scans):
            # Add new positioners and detectors to the existing lists
            positioners = (
                scan.positioners
                if positioners is None
                else positioners + scan.positioners
            )
            detectors = (
                scan.detectors if detectors is None else detectors + scan.detectors
            )

            # Check 'multi-channel-analyser-spectra of fluorescence-detector' names are as expected
            if i == 1:
                assert scan.labels() == DanteFluorescence.FLUOR_NAMES, (
                    f"Fluorescence detector labels do not match expected values. Found {scan.labels()}"
                )

            scan_labels = labels[i]
            scan_units = units[i]
            for p in scan.positioners:
                scan_labels.append(p.name)
                scan_units.append(p.unit)
            for d in scan.detectors:
                scan_labels.append(d.name)
                scan_units.append(d.unit)

        if len(mda_scans) == 1:
            labels = labels[0]
            units = units[0]

        if mda_arrays is not None:
            data = tuple(mda_arrays)

        # Collect a full set of the column descriptions
        column_types = {
            "Positioner": (
                "name",
                "descr",
                "step mode",
                "unit",
                "rdbk name",
                "rdbk descr",
                "rdbk unit",
            ),
            "Detector": (
                "name",
                "descr",
                "unit",
            ),
        }

        column_descriptions = (
            {
                i: [
                    p.name,
                    p.desc,
                    p.step_mode,
                    p.unit,
                    p.readback_name,
                    p.readback_desc,
                    p.readback_unit,
                ]
                for i, p in enumerate(positioners)
            }
            if positioners is not None
            else {}
        )
        if detectors is not None:
            column_descriptions.update(
                {
                    i + (len(positioners) if positioners is not None else 0): [
                        d.name,
                        d.desc,
                        d.unit,
                    ]
                    for i, d in enumerate(detectors)
                }
            )
        params["column_types"] = column_types
        params["column_descriptions"] = column_descriptions

        # Use scan time if available, otherwise let system time be used.
        if "MEX1ES01GLU01:MEX_TIME" in params:
            assert (
                params["MEX1ES01GLU01:MEX_TIME"][0] == "Seconds since 1990.01.01"
            )  # HAHA this is actually 2023.01.01 :')
            newtime = MEX2_datetime_conversion(params["MEX1ES01GLU01:MEX_TIME"][2])
            params["created"] = newtime

        return data, labels, units, params

    # @typing.overrides.typing.overrides # type: ignore[typing.override]
    @parserBase.summary_param_names_with_units.getter
    def summary_params_with_units(self) -> list[str]:
        """
        A dictionary of summary parameters with units.

        Returns
        -------
        list[str]
            A list of summary parameters, with interpereted units attached.
        """
        return [
            (
                f"{self.params[param][2]} ({self.params[param][1]})"
                if isinstance(self.params[param], tuple)
                and len(self.params[param]) == 3
                else self.params[param][0]
            )
            for param in self.SUMMARY_PARAM_RAW_NAMES
            if param in self.params
        ]


def MEX2_to_QANT_AUMainAsc(
    parser: parserBase,
    extrainfo_mapping: dict[str, None | str] = {
        "SR14ID01MCS02FAM:X.RBV": None,
        "SR14ID01MCS02FAM:Y.RBV": None,
        "SR14ID01MCS02FAM:Z.RBV": None,
        "SR14ID01MCS02FAM:R1.RBV": None,
        "SR14ID01MCS02FAM:R2.RBV": None,
        "SR14ID01NEXSCAN:saveData_comment1": "Sample",
        # "SR14ID01NEXSCAN:saveData_comment1": "MEX2SSCAN01:saveData_comment1",
        "SR14ID01NEXSCAN:saveData_comment2": None,
    },
) -> list[str]:
    """
    Convert a parser mapping to the QANT format readable.

    Parameters
    ----------
    parser : parserBase
        The parser object (with data, labels, units, and params loaded) to convert.
    extrainfo_mapping : dict[str, str | None], optional
        Optional mapping for known read-values for the QANT AUMainAsc format to
        parser parameter names. By default the dictionary key values (readable by QANT) are:
            {"SR14ID01MCS02FAM:X.RBV": None,
            "SR14ID01MCS02FAM:Y.RBV": None,
            "SR14ID01MCS02FAM:Z.RBV": None,
            "SR14ID01MCS02FAM:R1.RBV": None,
            "SR14ID01MCS02FAM:R2.RBV": None,
            "SR14ID01NEXSCAN:saveData_comment1": None,
            "SR14ID01NEXSCAN:saveData_comment2": None,}

    Returns
    -------
    list[str]
        A list of lines for the QANT AUMainAsc format, with newline terminations included.
    """
    import datetime as dt
    from typing import Hashable

    possible_read_values = [
        "SR14ID01MCS02FAM:X.RBV",
        "SR14ID01MCS02FAM:Y.RBV",
        "SR14ID01MCS02FAM:Z.RBV",
        "SR14ID01MCS02FAM:R1.RBV",
        "SR14ID01MCS02FAM:R2.RBV",
        "SR14ID01NEXSCAN:saveData_comment1",
        "SR14ID01NEXSCAN:saveData_comment2",
    ]
    # Check validity of the extrainfo_mapping
    parser.relabel = (
        True  # Ensure relabel is set to True, as this is needed for the conversion.
    )
    for key, value in extrainfo_mapping.items():
        if value is not None:
            if value not in parser.params:
                raise ValueError(f"Parameter {value} not found in parser params.")
            elif key not in possible_read_values:
                raise ValueError(f"Parameter {key} not found in possible read values.")

    # Create reverse dict
    extrainfo_remapping = {}
    for k, v in extrainfo_mapping.items():
        if v is None:
            continue
        if v in extrainfo_remapping:
            raise ValueError(
                f"Value {v} already in remapping - conflicting mapping for `{extrainfo_remapping[v]}` and `{k}`."
            )
        extrainfo_remapping[v] = k

    # Check vailidty of parser object:
    if parser.data is None:
        raise ValueError("Parser object does not have data loaded.")

    # Check validity of the dimensionality of the object:
    if len(parser.data.shape) != 2:
        raise ValueError("Parser object data is not 2D.")

    # Define a container for the output strings, line by line.
    ostrs = []

    # Define fake asc version number:
    ostrs.append("## mda2ascii 0.3.2 generated output\n")
    ostrs.append("\n")
    ostrs.append("\n")

    # Rename for consistency between mda and asc formats.
    mda_param_names = ["mda_version", "mda_scan_number", "mda_rank", "mda_dimensions"]
    asc_param_names = [
        "MDA File Version",
        "Scan Number",
        "Overall scan dimension",
        "Total requested scan size",
    ]
    for mda, asc in zip(mda_param_names, asc_param_names):
        if mda in parser.params:
            parser.params[asc] = parser.params[mda]
            del parser.params[mda]

    # Define MDA versioning from parameters, or create fictitious version.
    ostrs.append(
        "# MDA File Version = 1.3\n"
        if "MDA File Version" not in parser.params
        else f"# MDA File Version = {parser.params['MDA File Version']}\n"
    )
    ostrs.append(
        "# Scan Number = 1\n"
        if "Scan Number" not in parser.params
        else f"# Scan Number = {parser.params['Scan Number']}\n"
    )
    ostrs.append(
        "# Overall scan dimension = 1-D\n"
        if "Overall scan dimension" not in parser.params
        else f"# Overall scan dimension = {parser.params['Overall scan dimension']}-D\n"
    )
    ostrs.append(
        f"# Total requested scan size = {len(parser.data)}\n"
        if "Total requested scan size" not in parser.params
        else f"# Total requested scan size = {parser.params['Total requested scan size']}\n"
    )
    ostrs.append("\n")
    ostrs.append("\n")

    # Define the extra PVs
    ostrs.append("#  Extra PV: name, descr, values (, unit)\n")
    ostrs.append("\n")
    param_idx = 1
    for param in parser.params:
        if param not in asc_param_names:
            wparam = (
                param
                if param not in extrainfo_remapping
                or extrainfo_remapping[param] is None
                else extrainfo_remapping[param]
            )
            line = f"# Extra PV {param_idx}: {wparam}"
            if (
                hasattr(parser.params[param], "__len__")
                and type(parser.params[param]) is not str
            ):
                for val in parser.params[param]:
                    wval = (
                        val
                        if not isinstance(val, Hashable)
                        or val not in extrainfo_remapping
                        or extrainfo_remapping[val] is None
                        else extrainfo_remapping[val]
                    )
                    line += f", {wval}"
                line += "\n"
            else:
                line += f", {parser.params[param]}, \n"
            ostrs.append(line)
            param_idx += 1
    ostrs.append("\n")
    ostrs.append("\n")

    # Define the scan header:
    ostrs.append("# 1-D Scan\n")
    ostrs.append(
        f"# Points completed = {parser.data.shape[0]} of {parser.data.shape[0]}\n"
        if "Points completed" not in parser.params
        else f"# Points completed = {parser.params['Points completed']}\n"
    )
    ostrs.append(
        "# Scanner = SR14ID01NEXSCAN:scan1\n"
        if "Scanner" not in parser.params
        else f"# Scanner = {parser.params['Scanner']}\n"
    )
    if "Scan time" in parser.params:
        ostrs.append(f"# Scan time = {parser.params['Scan time']}\n")
    elif "created" in parser.params and isinstance(
        parser.params["created"], dt.datetime
    ):
        ostrs.append(
            f"# Scan time = {parser.params['created'].strftime(r'%b %d, %Y %H:%M:%S.%f')}\n"
        )
    elif "modified" in parser.params and isinstance(
        parser.params["modified"], dt.datetime
    ):
        ostrs.append(
            f"# Scan time = {parser.params['modified'].strftime(r'%b %d, %Y %H:%M:%S.%f')}\n"
        )
    else:
        # Use current time
        ostrs.append(
            f"# Scan time = {dt.datetime.now().strftime(r'%b %d, %Y %H:%M:%S.%f')}\n"
        )
    ostrs.append("\n")

    if "column_types" in parser.params and isinstance(
        parser.params["column_types"], dict
    ):
        for coltype in parser.params["column_types"].keys():
            line = f"#  {coltype}:"
            for val in parser.params["column_types"][coltype]:
                line += f" {val}"
                if val != parser.params["column_types"][coltype][-1]:
                    line += ","
                else:
                    line += "\n"
            ostrs.append(line)
    else:
        # Use default
        ostrs.append(
            "#  Positioner: name, descr, step mode, unit, rdbk name, rdbk descr, rdbk unit\n"
        )
        ostrs.append("#  Detector: name, descr, unit\n")
        ostrs.append("\n")

    # Define the Column Descriptions
    ostrs.append("# Column Descriptions:\n")
    if "column_descriptions" in parser.params:
        # Column descriptions have been saved. Use these.
        init_idx = 1  # default, add 1 to column description indexes
        if parser.params["column_descriptions"][0][0] != "Index":
            # Create index column if not present.
            ostrs.append(f"#{1:5}  [     Index      ]\n")
            init_idx = 2  # Because of added extra index.
        for i, col in parser.params["column_descriptions"].items():
            if "column_type_assignments" in parser.params:
                col_type = parser.params["column_type_assignments"][i]
            else:
                col_type = f"1-D Detector{i + 1:4}"  # Nth detector
            line = f"#{i + init_idx + 1:5}  [" + col_type + "]  "
            for val in col:
                line += f"{val}" if val is not None else ""
                if val != col[-1]:
                    line += ", "
                else:
                    line += "\n"
            ostrs.append(line)
    else:
        # Create column descriptions from units and labels.
        ostrs.append(f"#{1:5}  [     Index      ]\n")
        for i, label in enumerate(parser.labels):
            line = f"#{i + 2:5}  "
            if i == 0:
                # Assume energy labels...
                line += f"[1-D Positioner 1]  {label}, "
                if "units" in parser.params and parser.params["units"][i] is not None:
                    unit = parser.params["units"][i]
                    line += f"Mono setpoint, TABLE, {unit}, {label}, Mono setpoint, {unit}\n"
                else:
                    line += f"Mono setpoint, TABLE, eV, {label}, Mono setpoint, eV\n"
            else:
                # Assume detector labels, no description.
                line += f"[1-D Detector{i:4}]  {label}, "
                if "units" in parser.params and parser.params["units"][i] is not None:
                    unit = parser.params["units"][i]
                    line += f"{label}, , {unit}\n"
                else:  # no unit.
                    line += f"{label}, , \n"
            ostrs.append(line)
    ostrs.append("\n")

    # Define the Scan Values
    ostrs.append("# 1-D Scan Values\n")
    for i, row in enumerate(parser.data):
        line = f"{i + 1}"
        for val in row:
            line += f"\t{val}"
        line = line[:-1] + "\n"
        ostrs.append(line)

    # End of file
    return ostrs


if __name__ == "__main__":
    # Example usage
    path = os.path.dirname(__file__)
    package_path = os.path.normpath(os.path.join(path, "../../../../"))
    mda_paths = [
        os.path.normpath(
            os.path.join(
                package_path, f"tests/test_data/au/MEX2/2024-03/MEX2_564{i}.mda"
            )
        )
        for i in range(3)
        if i != 1
    ]
    mda_path1, mda_path2 = mda_paths
    # mda_path1, mda_path2, mda_path3 = mda_paths
    # mda_path1, mda_path2, mda_path3, mda_path4, mda_path5 = mda_paths
    print(mda_path1)
    print(mda_path2)
    # HEADER
    test1 = MEX2_NEXAFS(mda_path1, header_only=True)
    # BODY
    test2 = MEX2_NEXAFS(mda_path1, header_only=False)

    # Convert to a scan.
    test1.load()
    test1 = test1.to_scan(energy_bin_domain=(2230, 2390))
    test2 = test2.to_scan(load_all_columns=True)

    # Check that the domain can be manually applied.
    test3 = MEX2_NEXAFS(
        mda_path2,
        header_only=False,
    )

    # Check if previous binning is applied to new data.

    import matplotlib.pyplot as plt

    plt.close("all")
    subplts = plt.subplots(1, 1)
    fig: plt.Figure = subplts[0]
    ax: plt.Axes = subplts[1]
    idx = -1
    # ax.plot(test2.data[0][:, 0], test2.data[0][:, idx], label="Test2 " + test2.labels[0][idx])
    # [
    #     ax.plot(
    #         test.data[0][:, 0], test.data[0][:, idx], label=f"Tests[{i}]" + test.labels[0][idx]
    #     )
    #     for i, test in enumerate(tests)
    # ]
    # ax.plot(test3.data[0][:, 0], test3.data[0][:, idx], label="Test3 " + test3.labels[0][idx])
    ax.legend()
    # plt.ioff()

    plt.ion()
    # plt.show(block=False)
    plt.show(block=True)



from PyQt6 import QtWidgets
from PyQt6.QtWidgets import (
    QWidget,
    QLineEdit,
    QHBoxLayout,
    QVBoxLayout,
    QComboBox,
    QApplication,
    QLabel,
    QTextEdit,
    QCheckBox,
    QSplitter,
    QStyle,
    QProgressBar,
    QPushButton,
)

# from PyQt6.QtWidgets import QScrollBar, QHeaderView, QMainWindow, QTableWidget, QFrame, QGridLayout, QSizeGrip
from PyQt6.QtCore import (
    Qt,
    pyqtSignal,
)
import sys
from pyNexafs.parsers import parser_loaders, parserBase
from typing import Type
from pyNexafs.gui.widgets.io.dir_selection import directorySelector
from pyNexafs.gui.widgets.io.fileviewer import (
    directoryViewerTableNew,
    SummaryParamSelectorDialog,
)


class nexafsFileLoader(QWidget):
    """
    Header-viewer and file loader for NEXAFS data.

    Includes a variety of elements for selecting a directory, parser,
    filetype filters and header-contents filters. Filtering and sorting
    is implemented through QSortFilterProxyModel.


    Attributes
    ----------
    selectionLoaded : pyqtSignal
        Signal for selection change and completion of consequent file loading.
        Does not trigger if selection changes while no parser is selected.
    """

    selectionLoaded = pyqtSignal(bool)
    """A signal to determine when the """

    relabelling = pyqtSignal(bool)
    """A signal for changes if smart `relabelling` is to be applied to field names.
    This can be specified by the parsing classes."""

    def __init__(self, parent=None):
        super().__init__(parent)
        ## Instance attributes
        self._log_text = ""

        ## Initialise elements
        self.directory_selector = directorySelector(parent=self)
        self._progress_bar = QProgressBar()
        self.directory_viewer = directoryViewerTableNew(
            parent=self, progress_bar=self._progress_bar
        )
        self.nexafs_parser_selector = nexafsParserSelector()
        self.filter_relabelling = QCheckBox()
        self.filter = directoryFilterWidget(parent=self)
        self.log = QTextEdit()

        # Initalise graphics for loaded / unloaded files.
        self._icon_gear = QApplication.style().standardIcon(
            QStyle.StandardPixmap.SP_FileDialogDetailedView
        )
        self._select_params_btn = QPushButton(icon=self._icon_gear, text=None)
        self._select_params_btn.setToolTip("Select parser displayed parameters")

        ## Subplayouts
        dir_parser_layout = QHBoxLayout()
        dir_parser_layout.addWidget(self.directory_selector)
        dir_parser_layout.addWidget(QLabel("Parser:"))
        dir_parser_layout.addWidget(self.nexafs_parser_selector)
        dir_parser_layout.addWidget(QLabel("Relabel:"))
        dir_parser_layout.addWidget(self.filter_relabelling)

        filter_layout = QHBoxLayout()
        filter_layout.addWidget(self.filter)
        filter_layout.addWidget(self._select_params_btn)

        ## Qsplitter for expandable log
        draggable = QSplitter(Qt.Orientation.Vertical)
        draggable.addWidget(self.directory_viewer)
        draggable_log_widget = QWidget()
        draggable_log = QVBoxLayout()
        draggable_log.addWidget(QLabel("Log:"))
        draggable_log.addWidget(self.log)
        draggable_log_widget.setLayout(draggable_log)
        draggable_log_widget.setContentsMargins(0, 0, 0, 0)
        draggable.addWidget(draggable_log_widget)

        ## Dir loading progress bar
        self._progress_bar.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.MinimumExpanding,
            QtWidgets.QSizePolicy.Policy.Minimum,
        )
        self._progress_bar.setFixedHeight(15)
        self._progress_bar.setTextVisible(False)

        ## Directory Layout
        dir_layout = QVBoxLayout()
        dir_layout.addLayout(dir_parser_layout)
        dir_layout.addLayout(filter_layout)
        dir_layout.addWidget(self._progress_bar)
        dir_layout.addWidget(draggable)

        # Assign viewer layout to widget.
        self.setLayout(dir_layout)
        if parent is not None:
            self.setContentsMargins(0, 0, 0, 0)
            dir_layout.setContentsMargins(0, 0, 0, 0)

        ## Element attributes
        self.log.setReadOnly(True)
        self.directory_viewer.directory = (
            self.directory_selector._folder_path
        )  # Match initial value.
        draggable.setStretchFactor(0, 5)
        draggable.setStretchFactor(1, 1)

        ## Element Connections
        # Reload directory upon path change.
        self.directory_selector.new_path.connect(self.update_dir)
        # Reload directory upon parser change.
        self.nexafs_parser_selector.currentIndexChanged.connect(self.update_parser)
        # Reload directory upon filter change.
        self.filter.filterChanged.connect(self.update_filters)
        # Update relabelling
        self.filter_relabelling.stateChanged.connect(self.update_relabelling)
        # Signal for selection change.
        self.directory_viewer.selection_loaded.connect(self.on_selection_loaded)
        # Signal for selecting parser parameters.
        self._select_params_btn.clicked.connect(self.on_select_params)

        # Initialize directory.
        self._log_entry()

    def on_select_params(self):
        # Create a SummaryParamsSelectionDialog
        dialog = SummaryParamSelectorDialog(
            parsers=list(self.loaded_parser_headers.values()),
            parent=None,
            existing_selection=self.directory_viewer.header,
        )
        result = dialog.exec()
        if result:
            self.directory_viewer.header = sorted(list(dialog.selection))

    def on_selection_loaded(self):
        """
        Signal for selection change.
        """
        self.selectionLoaded.emit(True)
        return

    def update_relabelling(self) -> None:
        """
        Updates the fieldnames when relabelling is checked.
        """
        parser = self.directory_viewer.parser
        if parser is not None:
            old_relabel = parser.relabel
            new_relabel = self.filter_relabelling.isChecked()
            if old_relabel != new_relabel:
                parser.relabel = new_relabel
                self.directory_viewer.update_header()
                self.relabelling.emit(parser.relabel)

    def update_dir(self) -> None:
        """
        Update the dir selection
        """
        new_dir = self.directory_selector._folder_path
        # Update directory viewer
        self.directory_viewer.directory = new_dir
        # Log file updates.
        self._log_entry()
        return

    def update_parser(self) -> None:
        """
        Update the parser selection.
        """
        new_parser = self.nexafs_parser_selector.current_parser
        # Disable filter signalling while parser is changed.
        self.filter._changing_parser = True
        # Update directory viewer and filter parser selection
        self.filter.parser = new_parser
        self.directory_viewer.parser = new_parser
        self.filter._changing_parser = False
        self._log_entry()
        return

    def update_filters(self):
        """
        Update the directory viewer.
        """
        # Push an update to the directory viewer.
        self.directory_viewer.filters = (
            self.filter.filter_text,
            self.filter.filetypes_selection,
        )
        self._log_entry()
        return

    def _log_entry(self):
        # Log the current selection
        self._log_selection()
        # Log the number of files.
        self._log_files()

    def _log_selection(self):
        """
        Logs the path, parser and filter selection.
        """
        # Log path selection
        path = self.directory_selector._folder_path
        self.log_text += (
            f"Path '{path}' loaded"
            if len(path) < 30
            else f"Path '{path[:13]}...{path[-13:]}' loaded"
        )

        # Log parser selection
        parser = self.nexafs_parser_selector.current_parser  # current parser selection
        if parser is not None:
            self.log_text += f" using parser '{parser.__name__}'"

        # Log Filetype Selection?
        filetypes = self.filter.filetypes_selection
        if len(filetypes) == 1:
            self.log_text += f" with filetype '{filetypes[0]}'"

        # Log Filter Selection
        filter_text = self.filter.filter_text
        if filter_text != "" and filter_text is not None:
            self.log_text += f" with filter '{filter_text}'"

        # End line
        self.log_text += ". "  # no whitespace, assuming log files will be called next.
        return

    def _log_files(self):
        """
        Logs the number of files located after updating parameters.
        """
        file_count = self.directory_viewer.verticalHeader().count()  # filtered viewport
        dv_f = self.directory_viewer.files
        dv_ph = self.directory_viewer._parser_headers
        dv_p = self.directory_viewer.parser
        # Loading method is phrased differently depending on if a parser is selected.
        method = "parsed" if dv_p is not None and len(dv_ph) >= 0 else "located"
        # Determine endpoint loading.
        if file_count != 0 and len(dv_f) > 0:  # Filtered files loaded
            self.log_text += f"{file_count} {method} files.\n"
        elif file_count != 0 and file_count > 0:  # Filtered files
            self.log_text += f"Located {file_count} files.\n"
        else:
            self.log_text += "No files located.\n"

    @property
    def log_text(self):
        return self._log_text

    @log_text.setter
    def log_text(self, text: str):
        # Check if at max height
        sb = self.log.verticalScrollBar()
        atMaxHeight = True if sb.maximum() - sb.value() < 30 else False
        prevHeight = sb.value()
        # Update log
        self._log_text = text
        self.log.setText(text)
        # If was at max height, return to new max height.
        sb.setValue(sb.maximum() if atMaxHeight else prevHeight)

    @log_text.deleter
    def log_text(self):
        self._log_text = ""
        self.log.setText("")

    @property
    def loaded_parser_headers(self) -> dict[str, Type[parserBase] | None]:
        """
        Returns the parser headers loaded in the directory viewer.

        Returns
        -------
        dict[str, Type[parserBase]]
            A dictionary of filenames and their corresponding parser headers.
        """
        return self.directory_viewer._parser_headers

    @property
    def loaded_parsers_files(self) -> dict[str, Type[parserBase] | None]:
        """
        Returns the full-file parsers loaded in the directory viewer.

        Returns
        -------
        dict[str, Type[parserBase]]
            A dictionary of filenames and their corresponding parser objects.
        """
        return self.directory_viewer._parser_files

    @property
    def loaded_parser_files_selection(self) -> dict[str, Type[parserBase] | None]:
        """
        Returns the selection subset of full-file parsers from the directory viewer.

        Returns
        -------
        dict[str, Type[parserBase]]
            A dictionary of filenames and their corresponding scan objects.
        """
        selected = self.directory_viewer.selected_filenames
        loaded = self.loaded_parsers_files
        return {
            name: loaded[name]
            for name in selected
            if name in loaded and loaded[name] is not None
        }

    @property
    def selected_filenames(self) -> list[str]:
        """
        Returns the selected filenames in the directory viewer.

        Returns
        -------
        list[str]
            A list of filenames selected in the directory viewer.
        """
        return self.directory_viewer.selected_filenames


class nexafsParserSelector(QComboBox):
    """
    Selector for the type of NEXAFS data to load.
    """

    def __init__(self):
        super().__init__()

        if "" in parser_loaders:
            raise ValueError("Empty string is a reserved parser name.")

        # Define the relationships
        self.parsers = {"": None}
        self.parsers.update(parser_loaders)
        # Add items to the combobox.
        self.addItems([key for key in self.parsers.keys()])

    @property
    def current_parser(self) -> type[parserBase] | None:
        """
        Current selected parser.

        Returns
        -------
        type[parserBase] | None
            Returns the current parser which inherits from parserBase, or None if no parser is selected.
        """
        return self.parsers[self.currentText()]

    def update_combo_list(self):
        """
        Update the combo box parser list from the parser attribute.
        """
        self.clear()
        self.addItems([key for key in self.parsers.keys()])


class directoryFilterWidget(QWidget):
    """
    A widget for filtering the directory based on text and/or filetype.

    Generates a QTextEdit & QComboBox for filtering by text and/or filetype
    respectively. Default filetype options for a selected parser are populated
    by the parser.ALLOWED_EXTENSIONS constant.

    Attributes
    ----------
    filterChanged : pyqtSignal
        Emits upon a change in filter, useful for triggering table update.
    """

    filterChanged = pyqtSignal(bool)

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent=parent)

        layout = QHBoxLayout()
        self.setLayout(layout)

        # Update margins to be compact if a parent is provided.
        if parent is not None:
            self.setContentsMargins(0, 0, 0, 0)
            layout.setContentsMargins(0, 0, 0, 0)

        # Instance attributes
        self.filter_text = ""
        self.filter_filetype = None
        self._parser_selection = None
        self._changing_parser = False  # Flag to prevent filterChanged signal emitting when parser is changed.

        # Instance widgets
        self.filter_text_edit = QLineEdit()
        self.filter_filetype_select = QComboBox()

        # Setup layout
        layout.addWidget(self.filter_text_edit)
        layout.addWidget(QLabel("Filetype: "))
        layout.addWidget(self.filter_filetype_select)

        # Widget Attributes
        self.filter_text_edit.setPlaceholderText("Filter filename|header by text")

        # Connections
        self.filter_text_edit.editingFinished.connect(self.on_filter_text_edit)
        self.filter_filetype_select.currentIndexChanged.connect(self.on_filter_change)

    def on_filter_text_edit(self):
        """
        Update the filter text.
        """
        self.filter_text = self.filter_text_edit.text()
        self.on_filter_change()

    def on_filter_change(self):
        """
        Signal generator for filter changes if parser not changing.
        """
        if not self._changing_parser:
            self.filterChanged.emit(True)

    @property
    def filetypes_selection(self) -> list[str]:
        """
        Returns the currently selected filetype.

        Returns
        -------
        str
            The currently selected filetype, or a list of all possible filetypes, belonging to the parser.
            If parser is None, returns an empty list.
        """
        if self.parser is None:
            return []
        selection = self.filter_filetype_select.currentText()
        return [selection] if selection != "" else self.parser.ALLOWED_EXTENSIONS

    @property
    def parser(self) -> type[parserBase] | None:
        """
        Returns the currently selected parser.

        Returns
        -------
        type[parserBase] | None
            The currently used parser, or None if no parser is selected.
        """
        return self._parser_selection

    @parser.setter
    def parser(self, parser: type[parserBase] | None):
        """
        Sets the (external) parser selection.

        Additionally triggers the filetype selection to be updated and repopulates the filetype selection entries.

        Parameters
        ----------
        parser : type[parserBase] | None
            Currently used parser, or None if no parser is selected.
        """
        self._changing_parser = True
        self._parser_selection = parser
        self.filter_filetype_select.clear()  # clear existing file extensions.
        if parser is not None:
            self.filter_filetype_select.addItem("")
            self.filter_filetype_select.addItems(parser.ALLOWED_EXTENSIONS)
        self._changing_parser = False

    @parser.deleter
    def parser(self):
        """
        Removes the (external) parser selection.

        Additionally clears the filetype selections.
        """
        self._parser_selection = None
        self.filter_filetype_select.clear()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = nexafsFileLoader()
    window.show()
    window.setWindowTitle("pyNexafs File Loader")
    sys.exit(app.exec())



import numpy as np
from numpy.typing import NDArray


class FlourescenceDetector:
    """
    An abstract class to reduce energy-binned flourescence.

    Reduces multi-channel analyser (MCA) binned flourescence data.
    """

    FLUOR_NAMES: list[str]
    """The string names of the 2D flourescence binned channels (length N)."""
    BIN_ENERGY_DELTA: float
    """The energy difference (eV) between each bin"""
    TOTAL_BINS: int
    """The total number of energy bins in the channel data (length M)"""
    TOTAL_BIN_ENERGIES: NDArray
    """
    The energy (eV) of each channel bin.
    Should match the dimensions (M, N) where M is the number of bins, N is the number of channels.
    """
    BIN_ENERGY_OFFSET: float
    """The offset energy (eV) of the first energy bin."""
    INTERESTING_BIN_IDX: tuple[int, int] | None = None
    """
    Indexes of interesting energy bins from the detector.
    Can be used to hide energy bins with zero signal.
    """

    @classmethod
    def INTERESTING_BIN_ENERGIES(cls) -> NDArray:
        """
        The energies corresponding to the bins between the `INTERESTING_BIN_IDX` indices.

        Returns
        -------
        NDArray
            An array of the energies between the `INTERESTING_BIN_IDX` indices.

        Raises
        ------
        AttributeError
            When the class attribute `INTERESTING_BIN_IDX` doesn't have a defined tuple of indices.
        """
        if cls.INTERESTING_BIN_IDX is not None:
            return cls.TOTAL_BIN_ENERGIES[
                cls.INTERESTING_BIN_IDX[0] : cls.INTERESTING_BIN_IDX[1]
            ]
        else:
            raise AttributeError(
                f"`INTERESTING_BIN_ENERGIES` is not defined on the {cls.__name__} detector."
            )


class DanteFluorescence(FlourescenceDetector):
    """Configuration for interpreting the Dante MCA Fluorescence data."""

    FLUOR_NAMES: list[str] = [
        "MEX2ES01DPP01:ch1:W:ArrayData",
        "MEX2ES01DPP01:ch2:W:ArrayData",
        "MEX2ES01DPP01:ch3:W:ArrayData",
        "MEX2ES01DPP01:ch4:W:ArrayData",
    ]
    BIN_ENERGY_DELTA = 11.935
    BIN_ENERGY_OFFSET = 96 * 11.935 - 1146.7
    TOTAL_BINS = 4096
    TOTAL_BIN_ENERGIES = np.linspace(
        start=BIN_ENERGY_OFFSET,
        stop=BIN_ENERGY_OFFSET + TOTAL_BINS * BIN_ENERGY_DELTA,
        num=TOTAL_BINS,
    )
    INTERESTING_BIN_IDX = (80, 900)


class Xpress3Fluorescence(FlourescenceDetector):
    """Configuration for interpreting the Xpress3 MCA Fluorescence data."""

    FLUOR_NAMES = [
        "MEX2ES01DPP02:MCA1:ArrayData",  # New MCA
        "MEX2ES01DPP02:MCA2:ArrayData",
        "MEX2ES01DPP02:MCA3:ArrayData",
        "MEX2ES01DPP02:MCA4:ArrayData",
    ]
    BIN_ENERGY_DELTA = 10
    BIN_ENERGY_OFFSET = 0
    TOTAL_BINS = 4096
    TOTAL_BIN_ENERGIES = np.linspace(
        start=BIN_ENERGY_OFFSET,
        stop=TOTAL_BINS * BIN_ENERGY_DELTA,
        num=TOTAL_BINS,
    )
    INTERESTING_BIN_IDX = (0, 800)
